from __future__ import annotations

from math import exp, sqrt

import numpy as np
from numpy.linalg import norm

from dxtbx import flumpy
from libtbx.phil import parse
from scitbx import matrix
from scitbx.linalg import eigensystem, l_l_transpose_cholesky_decomposition_in_place

from dials.algorithms.profile_model.ellipsoid import (
    BBoxCalculatorAngular,
    BBoxCalculatorSimple,
    MaskCalculatorAngular,
    MaskCalculatorSimple,
    PredictorAngular,
    PredictorSimple,
)
from dials.algorithms.profile_model.ellipsoid.parameterisation import (
    Angular2MosaicityParameterisation,
    Angular4MosaicityParameterisation,
    Simple1MosaicityParameterisation,
    Simple6MosaicityParameterisation,
)
from dials.array_family import flex
from dials.constants import FULL_PARTIALITY
from dials.model.experiment.profile import ProfileModelExt

phil_scope = parse(
    """
rlp_mosaicity {

    model = simple1 simple6 angular2 *angular4
    .type = choice

}

wavelength_spread {

    model = *delta
    .type = choice

}

unit_cell {

    fixed = False
    .type = bool

}

orientation {

    fixed = False
    .type = bool

}

indexing {

    fail_on_bad_index = False
      .type = bool

  }

refinement {

    max_separation = 2
        .type = float

    outlier_probability = 0.975
        .type = float

    n_macro_cycles = 1
        .type = int

    n_cycles = 3
        .type = int

    min_n_reflections=10
        .type = int

}

prediction {
    d_min = None
        .type = float

    probability = %f
        .type = float
}"""
    % FULL_PARTIALITY
)


class EllipsoidProfileModel(ProfileModelExt):

    """
    An overall model class that conforms to the requirements of a
    dxtbx.profile_model entry point.
    """

    name = "ellipsoid"

    def __init__(self, parameterisation):
        self.parameterisation = parameterisation

    @classmethod
    def create(
        cls, params, reflections, crystal, beam, detector, goniometer=None, scan=None
    ):
        # a method to allow model creation for the standard integration program
        # need to work out the sigma d i.e. do initial integration
        from dials.algorithms.profile_model.gaussian_rs.calculator import (
            ComputeEsdBeamDivergence,
        )

        if not reflections:
            raise ValueError(
                "Reflections needed to determine sigma_d for the ellipsoid integrator"
            )

        sel = reflections.get_flags(reflections.flags.strong)
        strong_refls = reflections.select(sel)
        # Compute and initial spot size estimate and beam vector
        sigma_d = ComputeEsdBeamDivergence(detector, strong_refls).sigma()
        return cls.from_sigma_d(params.ellipsoid.rlp_mosaicity.model, sigma_d)

    def compute_bbox(
        self, reflections, crystal, beam, detector, goniometer=None, scan=None, **kwargs
    ):
        raise ValueError(
            "Ellipsoid profile modelling not implemented outside of dials.ssx_integrate"
        )

    @classmethod
    def from_sigma_d(cls, model, sigma_d):
        if model == "simple1":
            return cls(Simple1ProfileModel.from_sigma_d(sigma_d))
        elif model == "simple6":
            return cls(Simple6ProfileModel.from_sigma_d(sigma_d))
        elif model == "angular2":
            return cls(Angular2ProfileModel.from_sigma_d(sigma_d))
        elif model == "angular4":
            return cls(Angular4ProfileModel.from_sigma_d(sigma_d))

        raise RuntimeError(f"Unknown profile model: {model}")

    @classmethod
    def from_dict(cls, d):
        if d["parameterisation"] == "simple1":
            return cls(Simple1ProfileModel.from_dict(d))
        if d["parameterisation"] == "simple6":
            return cls(Simple6ProfileModel.from_dict(d))
        if d["parameterisation"] == "angular2":
            return cls(Angular2ProfileModel.from_dict(d))
        if d["parameterisation"] == "angular4":
            return cls(Angular4ProfileModel.from_dict(d))
        raise RuntimeError(
            f"Unknown profile model parameterisation: {d['parameterisation']}"
        )

    def mosaicity(self):
        return self.parameterisation.mosaicity()

    def sigma(self):
        return self.parameterisation.sigma()

    def to_dict(self):
        d = self.parameterisation.to_dict()
        d["parameterisation"] = d.pop("__id__")
        d["__id__"] = "ellipsoid"
        return d


class ProfileModelBase(object):
    """
    Class to store profile model

    """

    def __init__(self, params):
        """
        Initialise the class

        """
        self.params = params

    def sigma(self):
        """
        Get the sigma

        """
        return self.parameterisation().sigma()

    def update_model_state_parameters(self, state):
        """
        Update the model state with the parameters

        """
        state.set_M_params(self.params)

    def update_model(self, state):
        """
        Update the model

        """

        # Compute the eigen decomposition of the covariance matrix and check
        # largest eigen value
        sqr_mat = matrix.sqr(flumpy.from_numpy(self.sigma()))
        eigen_decomposition = eigensystem.real_symmetric(
            sqr_mat.as_flex_double_matrix()
        )
        L = eigen_decomposition.values()
        if L[0] > 1e-5:
            raise RuntimeError("Mosaicity matrix is unphysically large")

        self.params = state.M_params

    def to_dict(self):
        """Convert the model to a dictionary."""
        params = list(self.parameterisation().parameters)
        sigma = self.sigma()
        return {
            "__id__": self.__class__.name,
            "parameters": params,
            "sigma": sigma.tolist(),
        }

    @classmethod
    def from_dict(cls, d):
        """Convert the model to a dictionary."""
        model = cls(d["parameters"])
        return model

    def mosaicity(self):
        return self.parameterisation().mosaicity()


class SimpleProfileModelBase(ProfileModelBase):
    """
    Base class for simple profile models

    """

    def predict_reflections(
        self, experiments, miller_indices, probability=FULL_PARTIALITY
    ):
        """
        Predict the reflections

        """
        predictor = PredictorSimple(
            experiments[0], matrix.sqr(flumpy.from_numpy(self.sigma())), probability
        )
        return predictor.predict(miller_indices)

    def compute_bbox(self, experiments, reflections, probability=FULL_PARTIALITY):
        """
        Compute the bounding box

        """
        calculator = BBoxCalculatorSimple(
            experiments[0], matrix.sqr(flumpy.from_numpy(self.sigma())), probability, 4
        )
        calculator.compute(reflections)

    def compute_mask(self, experiments, reflections, probability=FULL_PARTIALITY):
        """
        Compute the mask

        """
        calculator = MaskCalculatorSimple(
            experiments[0], matrix.sqr(flumpy.from_numpy(self.sigma())), probability
        )
        calculator.compute(reflections)

    def sigma_for_reflection(self, s0, r):
        """
        Get sigma for a reflections

        """
        return np.array(self.sigma()).reshape(3, 3)

    def compute_partiality(self, experiments, reflections):
        """
        Compute the partiality

        """
        s0 = np.array([experiments[0].beam.get_s0()], dtype=np.float64).reshape(3, 1)
        s0_length = norm(s0)
        num = reflections.size()

        # Compute the marginal variance for the 000 reflection
        S00 = experiments[0].crystal.mosaicity.sigma()[2, 2]
        partiality = flex.double(reflections.size())
        partiality_variance = flex.double(reflections.size())
        for k, s2_vec in enumerate(reflections["s2"]):
            s2 = np.array(list(s2_vec), dtype=np.float64).reshape(3, 1)
            sigma = experiments[0].crystal.mosaicity.sigma()
            R = compute_change_of_basis_operation(s0, s2)

            S = np.matmul(R, np.array(sigma).reshape(3, 3))
            S = np.matmul(S, R.T)
            mu = np.matmul(R, s2)

            mu_norm = mu / norm(mu)
            assert abs(1.0 - mu_norm.flatten()[2]) < 1e-7
            S22 = S[2, 2]
            mu2 = mu.flatten()[2]
            eps = s0_length - mu2
            var_eps = S22 / num  # FIXME Approximation
            partiality[k] = exp(-0.5 * eps * (1 / S22) * eps) * sqrt(S00 / S22)
            partiality_variance[k] = (
                var_eps * (eps**2 / (S00 * S22)) * exp(eps**2 / S22)
            )

        reflections["partiality"] = partiality
        reflections["partiality.inv.variance"] = partiality_variance

    @classmethod
    def from_params(Class, params):
        """
        Create the class from some parameters

        """
        return Class(params)


class Simple1ProfileModel(SimpleProfileModelBase):
    """
    Simple 1 profile model class

    """

    name = "simple1"

    def parameterisation(self):
        """
        Get the parameterisation

        """
        return Simple1MosaicityParameterisation(self.params)

    @classmethod
    def from_sigma_d(Class, sigma_d):
        """
        Create the profile model from sigma_d estimate

        """
        return Class.from_params(np.array([sigma_d], dtype=np.float64))

    @classmethod
    def from_sigma(Class, sigma):
        """
        Construct the profile model from the sigma

        """

        # Construct triangular matrix
        LL = flex.double()
        for j in range(3):
            for i in range(j + 1):
                LL.append(sigma[j * 3 + i])

        # Do the cholesky decomposition
        _ = l_l_transpose_cholesky_decomposition_in_place(LL)
        TINY = 1e-6  ###FIXME
        assert abs(LL[1] - 0) < TINY
        assert abs(LL[2] - LL[0]) < TINY
        assert abs(LL[3] - 0) < TINY
        assert abs(LL[4] - 0) < TINY
        assert abs(LL[5] - LL[0]) < TINY

        # Setup the parameters
        return Class.from_params(flex.double((LL[0],)))


class Simple6ProfileModel(SimpleProfileModelBase):
    """
    Class to store profile model

    """

    name = "simple6"

    def parameterisation(self):
        """
        Get the parameterisation

        """
        return Simple6MosaicityParameterisation(self.params)

    @classmethod
    def from_sigma_d(Class, sigma_d):
        """
        Create the profile model from sigma_d estimate

        """
        return Class.from_params(
            np.array([sigma_d, 0, sigma_d, 0, 0, sigma_d], dtype=np.float64)
        )

    @classmethod
    def from_sigma(Class, sigma):
        """
        Construct the profile model from the sigma

        """

        # Construct triangular matrix
        LL = flex.double()
        for j in range(3):
            for i in range(j + 1):
                LL.append(sigma[j * 3 + i])

        # Do the cholesky decomposition
        _ = l_l_transpose_cholesky_decomposition_in_place(LL)

        # Setup the parameters
        return Class.from_params(
            flex.double((LL[0], LL[1], LL[2], LL[3], LL[4], LL[5]))
        )


class AngularProfileModelBase(ProfileModelBase):
    """
    Class to store profile model

    """

    def sigma_for_reflection(self, s0, r):
        """
        Sigma for a reflection

        """
        Q = compute_change_of_basis_operation(s0, r)
        sigma = np.matmul(np.matmul(Q.T, np.array(self.sigma()).reshape(3, 3)), Q)
        return sigma

    def predict_reflections(
        self, experiments, miller_indices, probability=FULL_PARTIALITY
    ):
        """
        Predict the reflections

        """
        predictor = PredictorAngular(
            experiments[0], matrix.sqr(flumpy.from_numpy(self.sigma())), probability
        )
        return predictor.predict(miller_indices)

    def compute_bbox(self, experiments, reflections, probability=FULL_PARTIALITY):
        """
        Compute the bounding box

        """
        calculator = BBoxCalculatorAngular(
            experiments[0], matrix.sqr(flumpy.from_numpy(self.sigma())), probability, 4
        )
        calculator.compute(reflections)

    def compute_mask(self, experiments, reflections, probability=FULL_PARTIALITY):
        """
        Compute the mask

        """
        calculator = MaskCalculatorAngular(
            experiments[0], matrix.sqr(flumpy.from_numpy(self.sigma())), probability
        )
        calculator.compute(reflections)

    def compute_partiality(self, experiments, reflections):
        """
        Compute the partiality

        """
        s0 = np.array([experiments[0].beam.get_s0()], dtype=np.float64).reshape(3, 1)
        s0_length = norm(s0)
        num = reflections.size()
        sigma = experiments[0].crystal.mosaicity.sigma()
        sigma = np.array(sigma).reshape(3, 3)
        x, y, z = reflections["s2"].parts()
        s2 = np.array([x, y, z])

        r = s2 - s0
        Rs = compute_change_of_basis_operations(s0, s2)
        Qs = compute_change_of_basis_operations(s0, r)
        sigma_qs = np.einsum("mda,db,mbc->mac", Qs, sigma, Qs)  # Q.T * sigma * Q
        Ss = np.einsum("mil,mlj,mkj->mik", Rs, sigma_qs, Rs)  # R * sigma_q * R.T
        mus = np.einsum("mij,jm->mi", Rs, s2)
        eps = mus[:, 2] - s0_length
        eps2 = np.square(eps)
        S22 = Ss[:, 2, 2]
        partiality = np.exp(-0.5 * eps2 / S22)
        partiality_variance = eps2 * np.exp(eps2 / S22) / (num * S22)
        # shortened versions of the original calculations below:
        # var_eps = S22 / num  # FIXME Approximation
        # S00 = S22  # FIXME
        # partiality[k] = exp(-0.5 * eps * (1 / S22) * eps) * sqrt(S00 / S22)
        # partiality_variance[k] = (
        #    var_eps * (eps**2 / (S00 * S22)) * exp(eps**2 / S22)
        # )

        reflections["partiality"] = flumpy.from_numpy(partiality)
        reflections["partiality.inv.variance"] = flumpy.from_numpy(partiality_variance)

    @classmethod
    def from_params(Class, params):
        """
        Create the class from some parameters

        """
        return Class(params)


class Angular2ProfileModel(AngularProfileModelBase):
    """
    Class to store profile model

    """

    name = "angular2"

    def parameterisation(self):
        """
        Get the parameterisation

        """
        return Angular2MosaicityParameterisation(self.params)

    @classmethod
    def from_sigma_d(Class, sigma_d):
        """
        Create the profile model from sigma_d estimate

        """
        return Class.from_params(np.array([sigma_d, sigma_d], dtype=np.float64))

    @classmethod
    def from_sigma(Class, sigma):
        """
        Construct the profile model from the sigma

        """

        # Construct triangular matrix
        LL = flex.double()
        for j in range(3):
            for i in range(j + 1):
                LL.append(sigma[j * 3 + i])

        # Do the cholesky decomposition
        _ = l_l_transpose_cholesky_decomposition_in_place(LL)

        # Check the sigma is as we expect
        TINY = 1e-10
        assert abs(LL[1] - 0) < TINY
        assert abs(LL[2] - LL[0]) < TINY
        assert abs(LL[3] - 0) < TINY
        assert abs(LL[4] - 0) < TINY

        # Setup the parameters
        return Class.from_params(flex.double((LL[0], LL[5])))


class Angular4ProfileModel(AngularProfileModelBase):
    """
    Class to store profile model

    """

    name = "angular4"

    def parameterisation(self):
        """
        Get the parameterisation

        """
        return Angular4MosaicityParameterisation(self.params)

    @classmethod
    def from_sigma_d(Class, sigma_d):
        """
        Create the profile model from sigma_d estimate

        """
        return Class.from_params(
            np.array([sigma_d, 0, sigma_d, sigma_d], dtype=np.float64)
        )

    @classmethod
    def from_sigma(Class, sigma):
        """
        Construct the profile model from the sigma

        """

        # Construct triangular matrix
        LL = flex.double()
        for j in range(3):
            for i in range(j + 1):
                LL.append(sigma[j * 3 + i])

        # Do the cholesky decomposition
        _ = l_l_transpose_cholesky_decomposition_in_place(LL)

        # Check the sigma is as we expect
        TINY = 1e-10
        assert abs(LL[3] - 0) < TINY
        assert abs(LL[4] - 0) < TINY

        # Setup the parameters
        return Class.from_params(flex.double((LL[0], LL[1], LL[2], LL[5])))


class ProfileModelFactory(object):
    """
    Class to create profile models

    """

    @classmethod
    def from_sigma_d(Class, model, sigma_d):
        """
        Construct a profile model from an initial sigma estimate

        """
        return EllipsoidProfileModel.from_sigma_d(model, sigma_d)


def compute_change_of_basis_operations(s0, s2_array):
    s0 = s0.reshape(1, 3)
    s2 = s2_array.T
    assert s2.shape[1] == 3
    e1 = np.cross(s2, s0)
    e2 = np.cross(s2, e1)
    norm_e1 = norm(e1, axis=1).reshape(-1, 1)
    norm_e2 = norm(e2, axis=1).reshape(-1, 1)
    e1 /= norm_e1
    e2 /= norm_e2
    norm_s2 = norm(s2, axis=1).reshape(-1, 1)
    e3 = s2 / norm_s2
    n = e1.shape[0]
    R = np.hstack([e1, e2, e3]).reshape(n, 3, 3)
    return R


def compute_change_of_basis_operation(s0, s2):
    s2 = s2.flatten()
    s0 = s0.flatten()
    e1 = np.cross(s2, s0)
    e2 = np.cross(s2, e1)
    e1 /= norm(e1)
    e2 /= norm(e2)
    e3 = s2 / norm(s2)
    R = np.array([e1, e2, e3], dtype=np.float64)
    return R



from __future__ import annotations

import copy

from scitbx.array_family import flex


def slice_experiments(experiments, image_ranges):
    """
    :param experiments
    :type experiments: dxtbx.model.experiment_list.ExperimentList
    :param image_range:
    :type image_range: list of 2-tuples defining scan range for each experiment
    """

    # copy the experiments
    experiments = copy.deepcopy(experiments)

    if len(experiments) != len(image_ranges):
        raise ValueError(
            "Input experiment list and image_ranges are not of the same length"
        )

    for exp, sr in zip(experiments, image_ranges):
        if sr is None:
            continue
        im_range = exp.scan.get_image_range()
        if sr[0] < im_range[0] or sr[1] > im_range[1]:
            raise IndexError("requested slice outside current scan range")

        # slicing uses the array range, not the image range
        arr_start = exp.scan.get_array_range()[0]
        beg = sr[0] - 1 - arr_start
        end = sr[1] - arr_start
        exp.scan.swap(exp.scan[beg:end])
        if exp.imageset is not None:
            # Gorilla of temporary workarounds for inconsistent scan and imageset slicing
            # https://github.com/cctbx/dxtbx/issues/213
            offset = exp.scan.get_batch_offset()
            exp.imageset = exp.imageset[beg + offset : end + offset]

        # account for scan-varying crystal
        if exp.crystal and exp.crystal.num_scan_points > 0:
            exp.crystal = slice_crystal(exp.crystal, (beg, end))

    return experiments


def slice_reflections(reflections, image_ranges):
    """
    :param reflections: reflection table of input reflections
    :type reflections: dials.array_family.flex.reflection_table
    :param image_range: list of 2-tuples defining scan range for each experiment
                       id contained within the reflections
    :type image_range: list of 2-tuples defining scan range for each experiment
    """

    # copy the reflections
    reflections = copy.deepcopy(reflections)

    to_keep = flex.size_t()
    for iexp, sr in enumerate(image_ranges):

        if sr is None:
            continue
        isel = (reflections["id"] == iexp).iselection()
        frames = (reflections["xyzobs.px.value"].parts()[2]).select(isel)
        # reflns on image n have frames in range [n-1, n)
        in_low_lim = frames >= sr[0] - 1
        in_high_lim = frames < sr[1]
        in_lim = in_low_lim & in_high_lim

        # which indices to keep?
        sub_isel = isel.select(in_lim)
        to_keep.extend(sub_isel)

    # implicitly also removes any reflections with ID outside the range of the
    # length of image_ranges
    return reflections.select(to_keep)


def slice_crystal(crystal, array_range):
    """Slice a scan-varying crystal by the provided array range"""

    if crystal.num_scan_points == 0:
        return crystal
    crystal = copy.deepcopy(crystal)
    UB_mats = [crystal.get_A_at_scan_point(i) for i in range(crystal.num_scan_points)]
    beg, end = array_range
    end += 1  # one more scan-point than images in the range
    UB_mats = UB_mats[beg:end]
    B_cov = crystal.get_B_covariance_at_scan_points()

    crystal.reset_scan_points()
    crystal.set_A_at_scan_points(UB_mats)
    if len(B_cov) > 0:
        B_cov = B_cov[beg:end, 0:9, 0:9]
        crystal.set_B_covariance_at_scan_points(B_cov)

    return crystal



from __future__ import annotations

import logging
from math import isclose

from cctbx import crystal, miller
from iotbx.shelx.write_ins import LATT_SYMM

from dials.algorithms.scaling.scaling_library import determine_best_unit_cell
from dials.array_family import flex
from dials.util import Sorry
from dials.util.filter_reflections import filter_reflection_table

logger = logging.getLogger(__name__)


def export_shelx(scaled_data, experiment_list, params):
    """Export scaled data corresponding to experiment_list to
    a SHELX HKL formatted text file."""

    # Handle requesting profile intensities (default via auto) but no column
    if "profile" in params.intensity and "intensity.prf.value" not in scaled_data:
        raise Sorry(
            "Requested profile intensity data but only summed present. Use intensity=sum."
        )

    # use supplied best unit cell or that determined from experiment list to define d in reflection table.
    best_unit_cell = params.mtz.best_unit_cell
    if best_unit_cell is None:
        best_unit_cell = determine_best_unit_cell(experiment_list)
    else:
        logger.info("Using supplied unit cell across experiments : %s", best_unit_cell)
    scaled_data["d"] = best_unit_cell.d(scaled_data["miller_index"])

    # Clean up reflection table with mtz defaults (as in export_xds_ascii)
    scaled_data = filter_reflection_table(
        scaled_data,
        intensity_choice=params.intensity,
        partiality_threshold=params.mtz.partiality_threshold,
        combine_partials=params.mtz.combine_partials,
        min_isigi=params.mtz.min_isigi,
        filter_ice_rings=params.mtz.filter_ice_rings,
        d_min=params.mtz.d_min,
    )

    # Check that all experiments have the same space group
    if len({x.crystal.get_space_group().make_tidy() for x in experiment_list}) != 1:
        raise ValueError("Experiments do not have a unique space group")

    # Create miller set with space group from 1st crystal in experiment list and best unit cell
    miller_set = miller.set(
        crystal_symmetry=crystal.symmetry(
            unit_cell=best_unit_cell,
            space_group=experiment_list[0].crystal.get_space_group(),
        ),
        indices=scaled_data["miller_index"],
        anomalous_flag=False,
    )

    intensity_choice = (
        params.intensity[0] if params.intensity[0] != "profile" else "prf"
    )
    intensities, variances = (
        scaled_data["intensity." + intensity_choice + ".value"],
        scaled_data["intensity." + intensity_choice + ".variance"],
    )
    assert variances.all_gt(0)
    i_obs = miller.array(miller_set, data=intensities)
    i_obs.set_observation_type_xray_intensity()
    i_obs.set_sigmas(flex.sqrt(variances))

    # If requested scale up data to maximise use of scale_range
    if params.shelx.scale:
        max_val = max(
            i_obs.data().min_max_mean().max, i_obs.sigmas().min_max_mean().max
        )
        min_val = min(
            i_obs.data().min_max_mean().min, i_obs.sigmas().min_max_mean().min
        )
        min_scale = params.shelx.scale_range[0] / min_val
        max_scale = params.shelx.scale_range[1] / max_val
        scale = min(min_scale, max_scale)
        i_obs = i_obs.apply_scaling(factor=scale)

    # write the SHELX HKL file
    hkl_file = params.shelx.hklout
    with open(hkl_file, "w") as hkl_file_object:
        i_obs.export_as_shelx_hklf(
            file_object=hkl_file_object,
            scale_range=params.shelx.scale_range,
            normalise_if_format_overflow=True,
        )
    logger.info(f"Written {i_obs.size()} reflections to {hkl_file}")

    # and a stub of an .ins file with information from the .expt file
    _write_ins(
        experiment_list,
        best_unit_cell=params.mtz.best_unit_cell,
        ins_file=params.shelx.ins,
    )
    logger.info(f"Written {params.shelx.ins}")


def _write_ins(experiment_list, best_unit_cell, ins_file):
    sg = experiment_list[0].crystal.get_space_group()
    unit_cells = []
    wavelengths = []

    # Check for single wavelength
    for exp in experiment_list:
        wl = exp.beam.get_wavelength()
        if not any([isclose(wl, w, abs_tol=1e-4) for w in wavelengths]):
            wavelengths.append(wl)
    if len(wavelengths) > 1:
        raise ValueError("Experiments have more than one wavelength")
    else:
        wl = wavelengths[0]

    # if user has supplied best_unit_cell use it
    if best_unit_cell is not None:
        uc = best_unit_cell
        uc_sd = None
    else:
        for exp in experiment_list:
            unit_cells.append(
                exp.crystal.get_recalculated_unit_cell() or exp.crystal.get_unit_cell()
            )

        if len(unit_cells) > 1:
            if (
                len({uc.parameters() for uc in unit_cells}) > 1
            ):  # have different cells so no esds
                uc = determine_best_unit_cell(experiment_list)
                uc_sd = None
            else:  # identical (recalculated?) unit cell with esds
                uc = (
                    experiment_list[0].crystal.get_recalculated_unit_cell()
                    or experiment_list[0].crystal.get_unit_cell()
                )
                uc_sd = (
                    experiment_list[0].crystal.get_recalculated_cell_parameter_sd()
                    or experiment_list[0].crystal.get_cell_parameter_sd()
                )
        else:  # single unit cell
            uc = (
                experiment_list[0].crystal.get_recalculated_unit_cell()
                or experiment_list[0].crystal.get_unit_cell()
            )
            uc_sd = (
                experiment_list[0].crystal.get_recalculated_cell_parameter_sd()
                or experiment_list[0].crystal.get_cell_parameter_sd()
            )

    with open(ins_file, "w") as f:
        f.write(
            f"TITL {sg.type().number()} in {sg.type().lookup_symbol().replace(' ','')}\n"
        )
        f.write(
            "CELL {:8.5f} {:8.4f} {:8.4f} {:8.4f} {:8.3f} {:8.3f} {:8.3f}\n".format(
                wl, *uc.parameters()
            )
        )
        if uc_sd:
            f.write(
                "ZERR {:8.3f} {:8.4f} {:8.4f} {:8.4f} {:8.3f} {:8.3f} {:8.3f}\n".format(
                    sg.order_z(), *uc_sd
                )
            )
        LATT_SYMM(f, sg)



"""
Tests for dials.command_line.anvil_correction.
"""


from __future__ import annotations

import copy
from pathlib import Path

import pytest

from dxtbx.model import ExperimentList

from dials.array_family import flex
from dials.command_line.anvil_correction import (
    correct_intensities_for_dac_attenuation,
    run,
)


def test_correct_correction(dials_data):
    """Test that the anvil absorption correction is producing expected values."""
    data_dir = dials_data("centroid_test_data", pathlib=True)

    # We'll need an integrated reflection table and an experiment list.
    reflections_file = data_dir / "integrated.pickle"
    experiments_file = data_dir / "experiments.json"

    # We need only test with the first ten reflections.
    reflections = flex.reflection_table.from_file(reflections_file)
    reflections = reflections.select(flex.size_t_range(10))

    experiment = ExperimentList.from_file(experiments_file)[0]

    # Test the correction that would be applied to a DAC with 1.5mm-thick anvils,
    # aligned along the z-axis at goniometer zero-datum.
    old_reflections = copy.deepcopy(reflections)
    correct_intensities_for_dac_attenuation(experiment, reflections, (0, 0, 1), 1.5)

    cases = {
        "intensity.sum.value": reflections.flags.integrated_sum,
        "intensity.sum.variance": reflections.flags.integrated_sum,
        "intensity.prf.value": reflections.flags.integrated_prf,
        "intensity.prf.variance": reflections.flags.integrated_prf,
    }
    corrections = flex.double(
        [
            0,
            6.653068275094517,
            6.522657529202368,
            6.3865190053761,
            6.587270967838122,
            6.43403642876391,
            6.39216742203502,
            0,
            6.152148372872684,
            6.0474840161407375,
        ]
    )
    for case, flag in cases.items():
        flagged = reflections.get_flags(flag)

        target_correction = corrections.select(flagged)
        if "variance" in case:
            target_correction = flex.pow2(target_correction)

        intensity_correction = (reflections[case] / old_reflections[case]).select(
            flagged
        )

        # Check that the un-integrated reflections are unchanged.
        assert pytest.approx(reflections[case].select(~flagged)) == old_reflections[
            case
        ].select(~flagged), (
            "Un-integrated reflections have been erroneously " "'corrected'."
        )

        # Check that the applied corrections are correct.
        assert pytest.approx(intensity_correction, rel=1e-5) == list(
            target_correction
        ), f"The applied intensity correction to {case} doesn't seem to be correct."


def test_help_message(dials_data, capsys):
    """Test that we get a help message when improper input is provided."""
    data_dir = dials_data("centroid_test_data", pathlib=True)

    # We'll need an integrated reflection table and an experiment list.
    reflections_file = str(data_dir / "integrated.pickle")
    experiments_file = str(data_dir / "experiments.json")

    for arguments in (
        [],
        [reflections_file],
        [experiments_file],
        [experiments_file, reflections_file, "anvil.normal=0,0,0"],
    ):
        with pytest.raises(SystemExit):
            run(arguments)
            assert (
                "Usage: dials.anvil_correction [options] integrated.expt "
                "integrated.refl" in capsys.readouterr().err
            )


def test_command_line(dials_data, run_in_tmp_path):
    """Test that the program runs correctly."""
    data_dir = dials_data("centroid_test_data", pathlib=True)

    # We'll need an integrated reflection table and an experiment list.
    reflections_file = str(data_dir / "integrated.pickle")
    experiments_file = str(data_dir / "experiments.json")

    run([experiments_file, reflections_file])

    assert Path("corrected.refl").is_file()

    with Path("dials.anvil_correction.log").open() as f:
        logfile = f.read()

    assert "Correcting integrated reflection intensities" in logfile
    assert "Writing the reflection table" in logfile



from __future__ import annotations

from cctbx.array_family import flex
from iotbx.data_plots import table_data
from libtbx import phil

from dials.util import resolution_analysis
from dials_pychef_ext import ChefStatistics, Observations

__all__ = [
    "ChefStatistics",
    "Observations",
    "Statistics",
    "batches_to_dose",
    "dose_phil_str",
    "phil_scope",
    "remove_batch_gaps",
    "resolution_limit",
    "table_data",
]

dose_phil_str = """\
dose {
  remove_gaps = True
    .type = bool

  batch
    .multiple = True
  {
    range = None
      .type = ints(value_min=0, size=2)
    dose_start = None
      .type = float(value_min=0)
    dose_step = None
      .type = float(value_min=0)
  }

}
"""

phil_scope = phil.parse(
    """\
d_min = None
  .type = float(value_min=0)
d_max = None
  .type = float(value_min=0)
min_completeness = None
  .type = float(value_min=0, value_max=1)
  .help = "Minimum value of completeness in outer resolution shell used to "
          "determine suitable resolution cutoff for analysis"
resolution_bins = 8
  .type = int
anomalous = False
  .type = bool
range {
  width = 1
    .type = float(value_min=0)
  min = None
    .type = float(value_min=0)
  max = None
    .type = float(value_min=0)
}
%s
"""
    % dose_phil_str
)


def interpret_images_to_doses_options(
    experiments, dose_per_image, starting_doses=None, shared_crystal=False
):
    """Interpret the dose.experiments options"""
    if len(dose_per_image) == 1:
        doses_per_image = dose_per_image * len(experiments)
    elif len(dose_per_image) != len(experiments):
        raise ValueError(
            """
The dose_per_image option must provide either one value, or a number of values
equal to the number of experiments (%s)"""
            % len(experiments)
        )
    else:
        doses_per_image = dose_per_image

    if shared_crystal:
        start_doses = [0]
        accumulated_dose = 0
        # adjust starting doses to account for a shared crystal.
        for expt, dose_per_img in zip(experiments, doses_per_image):
            imgrange = expt.scan.get_image_range()
            n_images = imgrange[1] - imgrange[0] + 1
            accumulated_dose += n_images * dose_per_img
            start_doses.append(accumulated_dose)
        start_doses = start_doses[:-1]
    elif starting_doses:
        if len(starting_doses) != len(experiments):
            raise ValueError(
                """
The number of starting_doses must equal the number of experiments (%s)"""
                % len(experiments)
            )
        start_doses = starting_doses
    else:
        start_doses = [0] * len(experiments)

    return start_doses, doses_per_image


class Statistics:
    def __init__(
        self, intensities, dose, n_bins=8, range_min=None, range_max=None, range_width=1
    ):

        if isinstance(dose, flex.double):
            sorted_dose = flex.sorted(dose)
            dd = sorted_dose[1:] - sorted_dose[:-1]
            dd = dd.select(dd > 0)
            if len(dd):
                step_size = flex.min(dd)
                dose /= step_size
            dose = dose.iround()
            if flex.min(dose) == 0:
                dose += 1

        # fix for completeness > 1 if screw axes present
        intensities = intensities.customized_copy(
            space_group_info=intensities.space_group()
            .build_derived_reflection_intensity_group(
                anomalous_flag=intensities.anomalous_flag()
            )
            .info(),
            info=intensities.info(),
        )

        self.intensities = intensities
        self.dose = dose
        self.n_bins = n_bins
        self.range_min = range_min
        self.range_max = range_max
        self.range_width = range_width
        assert self.range_width > 0

        if self.range_min is None:
            self.range_min = flex.min(self.dose) - self.range_width
        if self.range_max is None:
            self.range_max = flex.max(self.dose)
        self.n_steps = 2 + int((self.range_max - self.range_min) - self.range_width)

        sel = (self.dose.as_double() <= self.range_max) & (
            self.dose.as_double() >= self.range_min
        )
        self.dose = self.dose.select(sel)

        self.intensities = self.intensities.select(sel)
        self.d_star_sq = self.intensities.d_star_sq().data()

        self.binner = self.intensities.setup_binner_d_star_sq_step(
            d_star_sq_step=(flex.max(self.d_star_sq) - flex.min(self.d_star_sq) + 1e-8)
            / self.n_bins
        )

        # self.dose /= range_width
        self.dose -= int(self.range_min)

        self.dose = flex.size_t(list(self.dose))

        binner_non_anom = intensities.as_non_anomalous_array().use_binning(self.binner)
        n_complete = flex.size_t(binner_non_anom.counts_complete()[1:-1])

        chef_stats = ChefStatistics(
            intensities.indices(),
            intensities.data(),
            intensities.sigmas(),
            intensities.d_star_sq().data(),
            self.dose,
            n_complete,
            self.binner,
            intensities.space_group(),
            intensities.anomalous_flag(),
            self.n_steps,
        )

        self.iplus_comp_bins = chef_stats.iplus_completeness_bins()
        self.iminus_comp_bins = chef_stats.iminus_completeness_bins()
        self.ieither_comp_bins = chef_stats.ieither_completeness_bins()
        self.iboth_comp_bins = chef_stats.iboth_completeness_bins()
        self.iplus_comp_overall = chef_stats.iplus_completeness()
        self.iminus_comp_overall = chef_stats.iminus_completeness()
        self.ieither_comp_overall = chef_stats.ieither_completeness()
        self.iboth_comp_overall = chef_stats.iboth_completeness()
        self.rcp_bins = chef_stats.rcp_bins()
        self.rcp = chef_stats.rcp()
        self.scp_bins = chef_stats.scp_bins()
        self.scp = chef_stats.scp()
        self.rd = chef_stats.rd()

    def completeness_vs_dose_str(self):

        anomalous = self.intensities.anomalous_flag()

        title = "Completeness vs. dose:"
        graph_names = ["Completeness", "Completeness in resolution shells"]

        if anomalous:
            column_labels = (
                ["Dose"]
                + [
                    "%.2f-%.2f(A)" % self.binner.bin_d_range(i + 1)
                    for i in range(self.n_bins)
                ]
                + ["I+", "I-", "I", "dI"]
            )
            column_formats = (
                ["%8.1f"]
                + ["%5.3f" for i in range(self.n_bins)]
                + ["%5.3f", "%5.3f", "%5.3f", "%5.3f"]
            )
            # graph_columns = [[0,1,2,3,4]]
            graph_columns = [
                [0] + list(range(self.n_bins + 1, self.n_bins + 5)),
                list(range(self.n_bins + 1)),
            ]
        else:
            column_labels = (
                ["Dose"]
                + [
                    "%.2f-%.2f(A)" % self.binner.bin_d_range(i + 1)
                    for i in range(self.n_bins)
                ]
                + ["I"]
            )
            column_formats = (
                ["%8.1f"] + ["%5.3f" for i in range(self.n_bins)] + ["%5.3f"]
            )
            graph_columns = [[0, self.n_bins + 1], list(range(self.n_bins + 1))]

        table_completeness = table_data(
            title=title,
            column_labels=column_labels,
            column_formats=column_formats,
            graph_names=graph_names,
            graph_columns=graph_columns,
        )
        for i in range(self.n_steps):
            if anomalous:
                row = (
                    [i * self.range_width + self.range_min]
                    + [self.ieither_comp_bins[i_bin, i] for i_bin in range(self.n_bins)]
                    + [
                        self.iplus_comp_overall[i],
                        self.iminus_comp_overall[i],
                        self.ieither_comp_overall[i],
                        self.iboth_comp_overall[i],
                    ]
                )
            else:
                row = (
                    [i * self.range_width + self.range_min]
                    + [self.ieither_comp_bins[i_bin, i] for i_bin in range(self.n_bins)]
                    + [self.ieither_comp_overall[i]]
                )
            table_completeness.add_row(row)

        return table_completeness.format_loggraph()

    def _thing_vs_dose_shells_str(self, title, thing, value_bins, values):
        column_labels = (
            ["Dose"]
            + [
                "%.2f-%.2f(A)" % self.binner.bin_d_range(i + 1)
                for i in range(self.n_bins)
            ]
            + [thing]
        )
        column_formats = ["%8.1f"] + ["%7.4f" for i in range(self.n_bins + 1)]
        graph_names = [thing, thing + ", in resolution shells"]
        graph_columns = [[0, self.n_bins + 1], list(range(self.n_bins + 1))]
        table = table_data(
            title=title,
            column_labels=column_labels,
            column_formats=column_formats,
            graph_names=graph_names,
            graph_columns=graph_columns,
        )
        for i in range(self.n_steps):
            row = (
                [i * self.range_width + self.range_min]
                + [value_bins[j, i] for j in range(self.binner.n_bins_used())]
                + [values[i]]
            )
            table.add_row(row)
        return table.format_loggraph()

    def rcp_vs_dose_str(self):
        return self._thing_vs_dose_shells_str(
            "Cumulative radiation damage analysis:", "Rcp(d)", self.rcp_bins, self.rcp
        )

    def scp_vs_dose_str(self):
        return self._thing_vs_dose_shells_str(
            "Normalised radiation damage analysis:", "Scp(d)", self.scp_bins, self.scp
        )

    def rd_vs_dose_str(self):
        title = "R vs. BATCH difference:"
        column_labels = ["BATCH", "Rd"]
        column_formats = ["%8.1f", "%5.3f"]
        graph_names = ["Rd"]
        graph_columns = [[0, 1]]

        table_rd = table_data(
            title=title,
            column_labels=column_labels,
            column_formats=column_formats,
            graph_names=graph_names,
            graph_columns=graph_columns,
        )
        for i in range(self.n_steps):
            row = [i * self.range_width + self.range_min, self.rd[i]]
            table_rd.add_row(row)

        return table_rd.format_loggraph()

    def to_dict(self):
        x = [i * self.range_width + self.range_min for i in range(self.n_steps)]
        scp_data = []
        rcp_data = []
        rd_data = []
        completeness_data = []

        scp_data.append(
            {
                "x": x,
                "y": list(self.scp),
                "type": "scatter",
                "name": "Scp overall",
                "line": {"width": 3},
            }
        )
        rcp_data.append(
            {
                "x": x,
                "y": list(self.rcp),
                "type": "scatter",
                "name": "Rcp overall",
                "line": {"width": 3},
            }
        )
        rd_data.append({"x": x, "y": list(self.rd), "type": "scatter", "name": "Rd"})

        anomalous = self.intensities.anomalous_flag()
        completeness_data.append(
            {
                "x": x,
                "y": list(self.ieither_comp_overall),
                "type": "scatter",
                "name": "I",
                "line": {"width": 3},
            }
        )
        if anomalous:
            completeness_data.append(
                {
                    "x": x,
                    "y": list(self.iboth_comp_overall),
                    "type": "scatter",
                    "name": "dI",
                    "line": {"width": 3},
                }
            )
            completeness_data.append(
                {
                    "x": x,
                    "y": list(self.iplus_comp_overall),
                    "type": "scatter",
                    "name": "I+",
                    "line": {"width": 3},
                }
            )
            completeness_data.append(
                {
                    "x": x,
                    "y": list(self.iminus_comp_overall),
                    "type": "scatter",
                    "name": "I-",
                    "line": {"width": 3},
                }
            )

        if self.binner.n_bins_used() > 1:
            for j in range(self.binner.n_bins_used()):
                bin_range_suffix = " (%.2f-%.2f A)" % self.binner.bin_d_range(j + 1)
                scp_data.append(
                    {
                        "x": x,
                        "y": list(self.scp_bins[j : j + 1, :].as_1d()),
                        "type": "scatter",
                        "name": "Scp" + bin_range_suffix,
                        "line": {"width": 1, "dash": "dot"},
                    }
                )
                rcp_data.append(
                    {
                        "x": x,
                        "y": list(self.rcp_bins[j : j + 1, :].as_1d()),
                        "type": "scatter",
                        "name": "Rcp" + bin_range_suffix,
                        "line": {"width": 1, "dash": "dot"},
                    }
                )

                completeness_data.append(
                    {
                        "x": x,
                        "y": list(self.ieither_comp_bins[j : j + 1, :].as_1d()),
                        "type": "scatter",
                        "name": "I" + bin_range_suffix,
                        "line": {"width": 1, "dash": "dot"},
                    }
                )

        d = {
            "scp_vs_dose": {
                "data": scp_data,
                "layout": {
                    "title": "Scp vs dose",
                    "xaxis": {"title": "Dose"},
                    "yaxis": {"title": "Scp", "rangemode": "tozero"},
                },
            },
            "rcp_vs_dose": {
                "data": rcp_data,
                "layout": {
                    "title": "Rcp vs dose",
                    "xaxis": {"title": "Dose"},
                    "yaxis": {"title": "Rcp", "rangemode": "tozero"},
                },
            },
            "rd_vs_batch_difference": {
                "data": rd_data,
                "layout": {
                    "title": "Rd vs batch difference",
                    "xaxis": {"title": "Batch difference"},
                    "yaxis": {"title": "Rd", "rangemode": "tozero"},
                },
            },
            "completeness_vs_dose": {
                "data": completeness_data,
                "layout": {
                    "title": "Completeness vs dose",
                    "xaxis": {"title": "Dose"},
                    "yaxis": {"title": "Completeness", "rangemode": "tozero"},
                },
            },
        }
        return d


def batches_to_dose(batches, params):
    if len(params.batch):
        dose = flex.double(batches.size(), -1)
        for batch in params.batch:
            start = batch.dose_start
            step = batch.dose_step
            for i in range(batch.range[0], batch.range[1] + 1):
                # inclusive range
                dose.set_selected(batches == i, start + step * (i - batch.range[0]))
        dose = dose.iround()
    elif params.remove_gaps:
        dose = remove_batch_gaps(batches)
    else:
        dose = batches
    return dose


def remove_batch_gaps(batches):
    perm = flex.sort_permutation(batches)
    new_batches = flex.int(batches.size(), -1)
    sorted_batches = batches.select(perm)
    curr_batch = -1
    new_batch = -1
    for i, b in enumerate(sorted_batches):
        if b != curr_batch:
            curr_batch = b
            new_batch += 1
        new_batches[perm[i]] = new_batch
    return new_batches


def resolution_limit(i_obs, min_completeness, n_bins):
    params = resolution_analysis.phil_defaults.extract().resolution
    params.nbins = n_bins
    r = resolution_analysis.Resolutionizer(i_obs, params)
    return r.resolution(
        resolution_analysis.metrics.COMPLETENESS, limit=min_completeness
    )



from __future__ import annotations

import copy
import logging
import os
import pickle
import random

import pytest

from cctbx import sgtbx
from dxtbx.model import Crystal, Experiment, ExperimentList
from dxtbx.serialize import load

from dials.array_family import flex


def test_accessing_invalid_key_throws_keyerror():
    table = flex.reflection_table()
    with pytest.raises(KeyError) as e:
        table["missing_key"]
    assert e.value.args[0] == "Unknown column 'missing_key'"


def test_reflection_table_behaves_like_a_python_dictionary():
    def identical(flex1, flex2):
        return (flex1 == flex2).all_eq(True)

    flex_A = flex.int([1])
    flex_B = flex.int([2])
    table = flex.reflection_table([("A", flex_A), ("B", flex_B)])
    assert len(table) == 1
    assert list(table.keys()) == ["A", "B"]

    items = list(table.items())
    assert len(items) == 2
    assert items[0][0] == "A" and (items[0][1] == flex_A).all_eq(True)
    assert items[1][0] == "B" and (items[1][1] == flex_B).all_eq(True)

    assert list(table) == ["A", "B"]

    assert "A" in table and "B" in table and "C" not in table
    assert (table["A"] == flex_A).all_eq(True)
    assert (table["B"] == flex_B).all_eq(True)

    assert table.get("A", -424242) == flex_A
    assert table.get("C", -424242) == -424242


def test_init():
    # test default
    table = flex.reflection_table()
    assert table.is_consistent()
    assert table.nrows() == 0
    assert table.ncols() == 0
    assert table.empty()

    # test with nrows
    table = flex.reflection_table(10)
    assert table.is_consistent()
    assert table.nrows() == 10
    assert table.ncols() == 0
    assert table.empty()

    # test with valid columns
    table = flex.reflection_table(
        [
            ("col1", flex.int(10)),
            ("col2", flex.double(10)),
            ("col3", flex.std_string(10)),
        ]
    )
    assert table.is_consistent()
    assert table.nrows() == 10
    assert table.ncols() == 3
    assert not table.empty()

    # test with invalid columns
    with pytest.raises(RuntimeError):
        _ = flex.reflection_table(
            [
                ("col1", flex.int(10)),
                ("col2", flex.double(20)),
                ("col3", flex.std_string(10)),
            ]
        )


def test_resizing():
    # Create a table with 2 empty columns
    table = flex.reflection_table()
    assert table.empty()
    table["col1"] = flex.int()
    table["col2"] = flex.double()
    assert table.nrows() == 0
    assert table.ncols() == 2
    assert not table.empty()
    assert "col1" in table
    assert "col2" in table
    assert "col3" not in table

    # Create a table with 2 columns and 10 rows
    table = flex.reflection_table()
    table["col1"] = flex.int(10)
    table["col2"] = flex.double(10)
    assert table.nrows() == 10
    assert table.ncols() == 2

    # Add an extra column with the wrong size (throw)
    with pytest.raises(RuntimeError):
        table["col3"] = flex.std_string(20)
    assert table.nrows() == 10
    assert table.ncols() == 2
    assert table.is_consistent()
    assert len(table["col1"]) == 10
    assert len(table["col2"]) == 10
    assert len(table) == table.size()

    # Resize the table (should resize all columns)
    table.resize(50)
    assert table.nrows() == 50
    assert table.ncols() == 2
    assert table.is_consistent()
    assert len(table["col1"]) == 50
    assert len(table["col2"]) == 50

    # Make the table inconsistent
    table["col1"].resize(40)
    assert not table.is_consistent()
    with pytest.raises(Exception):
        table.nrows()
    assert table.ncols() == 2

    # Clear the table
    table.clear()
    assert table.is_consistent()
    assert table.empty()
    assert table.nrows() == 0
    assert table.ncols() == 0


def test_delete():
    # Test del item
    table = flex.reflection_table()
    table["col1"] = flex.int([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    table["col2"] = flex.int([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    table["col3"] = flex.int([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    del table["col3"]
    assert table.is_consistent()
    assert table.nrows() == 10
    assert table.ncols() == 2
    assert "col3" not in table

    # Test del row
    del table[5]
    assert table.is_consistent()
    assert table.nrows() == 9
    assert table.ncols() == 2
    assert all(a == b for a, b in zip(list(table["col1"]), [0, 1, 2, 3, 4, 6, 7, 8, 9]))

    # Test del slice
    del table[0:10:2]
    assert table.is_consistent()
    assert table.nrows() == 4
    assert table.ncols() == 2
    assert all(a == b for a, b in zip(list(table["col1"]), [1, 3, 6, 8]))

    # Test del slice
    del table[:]
    assert table.is_consistent()
    assert table.nrows() == 0
    assert table.ncols() == 2


def test_row_operations():
    # The columns as lists
    c1 = list(range(10))
    c2 = list(range(10))
    c3 = ["a", "b", "c", "d", "e", "f", "g", "i", "j", "k"]

    # Create a table with some elements
    table = flex.reflection_table()
    table["col1"] = flex.int(c1)
    table["col2"] = flex.double(c2)
    table["col3"] = flex.std_string(c3)

    # Extend the table
    table.extend(table)
    c1 = c1 * 2
    c2 = c2 * 2
    c3 = c3 * 2
    assert table.nrows() == 20
    assert table.ncols() == 3
    assert table.is_consistent()
    assert all(a == b for a, b in zip(table["col1"], c1))
    assert all(a == b for a, b in zip(table["col2"], c2))
    assert all(a == b for a, b in zip(table["col3"], c3))

    # Append some rows to the table
    row = {"col1": 10}
    c1 = c1 + [10]
    c2 = c2 + [0]
    c3 = c3 + [""]
    table.append(row)
    assert table.nrows() == 21
    assert table.ncols() == 3
    assert table.is_consistent()
    assert all(a == b for a, b in zip(table["col1"], c1))
    assert all(a == b for a, b in zip(table["col2"], c2))
    assert all(a == b for a, b in zip(table["col3"], c3))

    row = {"col2": 11}
    c1 = c1 + [0]
    c2 = c2 + [11]
    c3 = c3 + [""]
    table.append(row)
    assert table.nrows() == 22
    assert table.ncols() == 3
    assert table.is_consistent()
    assert all(a == b for a, b in zip(table["col1"], c1))
    assert all(a == b for a, b in zip(table["col2"], c2))
    assert all(a == b for a, b in zip(table["col3"], c3))

    row = {"col1": 12, "col2": 12, "col3": "l"}
    c1 = c1 + [12]
    c2 = c2 + [12]
    c3 = c3 + ["l"]
    table.append(row)
    assert table.nrows() == 23
    assert table.ncols() == 3
    assert table.is_consistent()
    assert all(a == b for a, b in zip(table["col1"], c1))
    assert all(a == b for a, b in zip(table["col2"], c2))
    assert all(a == b for a, b in zip(table["col3"], c3))

    # Try inserting some rows
    row = {"col1": -1}
    c1.insert(5, -1)
    c2.insert(5, 0)
    c3.insert(5, "")
    table.insert(5, row)
    assert table.nrows() == 24
    assert table.ncols() == 3
    assert table.is_consistent()
    assert all(a == b for a, b in zip(table["col1"], c1))
    assert all(a == b for a, b in zip(table["col2"], c2))
    assert all(a == b for a, b in zip(table["col3"], c3))

    row = {"col1": -2, "col2": -3, "col3": "abc"}
    c1.insert(2, -2)
    c2.insert(2, -3)
    c3.insert(2, "abc")
    table.insert(2, row)
    assert table.nrows() == 25
    assert table.ncols() == 3
    assert table.is_consistent()
    assert all(a == b for a, b in zip(table["col1"], c1))
    assert all(a == b for a, b in zip(table["col2"], c2))
    assert all(a == b for a, b in zip(table["col3"], c3))

    # Try iterating through table rows
    for i in range(table.nrows()):
        row = table[i]
        assert row["col1"] == c1[i]
        assert row["col2"] == c2[i]
        assert row["col3"] == c3[i]

    # Trying setting some rows
    row = {"col1": 100}
    table[2] = row
    assert table[2]["col1"] == 100
    assert table[2]["col2"] == c2[2]
    assert table[2]["col3"] == c3[2]

    row = {"col1": 1000, "col2": 2000, "col3": "hello"}
    table[10] = row
    assert table[10]["col1"] == 1000
    assert table[10]["col2"] == 2000
    assert table[10]["col3"] == "hello"


def test_iteration():
    # The columns as lists
    c1 = list(range(10))
    c2 = list(range(10))
    c3 = ["a", "b", "c", "d", "e", "f", "g", "i", "j", "k"]

    # Create a table with some elements
    table = flex.reflection_table()
    table["col1"] = flex.int(c1)
    table["col2"] = flex.double(c2)
    table["col3"] = flex.std_string(c3)

    # Try iterating keys
    k = []
    for key in table.keys():
        k.append(key)
    assert len(k) == 3
    assert k.count("col1") == 1
    assert k.count("col2") == 1
    assert k.count("col3") == 1

    # Try iterating columns
    k = []
    c = []
    for key, col in table.cols():
        k.append(key)
        c.append(col)
    assert len(k) == 3
    assert k.count("col1") == 1
    assert k.count("col2") == 1
    assert k.count("col3") == 1

    # Try iterating rows
    for row1, row2 in zip(table.rows(), zip(c1, c2, c3)):
        assert row1["col1"] == row2[0]
        assert row1["col2"] == row2[1]
        assert row1["col3"] == row2[2]


def test_slicing():
    # The columns as lists
    c1 = list(range(10))
    c2 = list(range(10))
    c3 = ["a", "b", "c", "d", "e", "f", "g", "i", "j", "k"]

    # Create a table with some elements
    table = flex.reflection_table()
    table["col1"] = flex.int(c1)
    table["col2"] = flex.double(c2)
    table["col3"] = flex.std_string(c3)

    # Try forward slicing
    new_table = table[2:7:2]
    assert new_table.ncols() == 3
    assert new_table.nrows() == 3
    assert new_table.is_consistent()
    c11 = c1[2:7:2]
    c22 = c2[2:7:2]
    c33 = c3[2:7:2]
    assert all(a == b for a, b in zip(new_table["col1"], c11))
    assert all(a == b for a, b in zip(new_table["col2"], c22))
    assert all(a == b for a, b in zip(new_table["col3"], c33))

    # Try backward slicing
    new_table = table[7:2:-2]
    assert new_table.ncols() == 3
    assert new_table.nrows() == 3
    assert new_table.is_consistent()
    c11 = c1[7:2:-2]
    c22 = c2[7:2:-2]
    c33 = c3[7:2:-2]
    assert all(a == b for a, b in zip(new_table["col1"], c11))
    assert all(a == b for a, b in zip(new_table["col2"], c22))
    assert all(a == b for a, b in zip(new_table["col3"], c33))

    # Try setting forward slicing
    table[2:7:2] = new_table
    assert table.ncols() == 3
    assert table.nrows() == 10
    assert table.is_consistent()
    c1[2:7:2] = c11
    c2[2:7:2] = c22
    c3[2:7:2] = c33
    assert all(a == b for a, b in zip(table["col1"], c1))
    assert all(a == b for a, b in zip(table["col2"], c2))
    assert all(a == b for a, b in zip(table["col3"], c3))

    # Try setting backward slicing
    table[7:2:-2] = new_table
    assert table.ncols() == 3
    assert table.nrows() == 10
    assert table.is_consistent()
    c1[7:2:-2] = c11
    c2[7:2:-2] = c22
    c3[7:2:-2] = c33
    assert all(a == b for a, b in zip(table["col1"], c1))
    assert all(a == b for a, b in zip(table["col2"], c2))
    assert all(a == b for a, b in zip(table["col3"], c3))


def test_updating():
    # The columns as lists
    c1 = list(range(10))
    c2 = list(range(10))
    c3 = ["a", "b", "c", "d", "e", "f", "g", "i", "j", "k"]

    # Create a table with some elements
    table0 = flex.reflection_table()
    table1 = flex.reflection_table()
    table2 = flex.reflection_table()
    table1["col1"] = flex.int(c1)
    table1["col2"] = flex.double(c2)
    table2["col3"] = flex.std_string(c3)

    # Update from zero columns
    table0.update(table1)
    assert table0.is_consistent()
    assert table0.nrows() == 10
    assert table0.ncols() == 2

    # Update table1 with table2 columns
    table1.update(table2)
    assert table1.is_consistent()
    assert table1.nrows() == 10
    assert table1.ncols() == 3
    assert table2.is_consistent()
    assert table2.nrows() == 10
    assert table2.ncols() == 1

    # Update trable1 with invalid table
    c3 = ["a", "b", "c"]

    # Create a table with some elements
    table2 = flex.reflection_table()
    table2["col3"] = flex.std_string(c3)
    with pytest.raises(RuntimeError):
        table1.update(table2)

    assert table1.is_consistent()
    assert table1.nrows() == 10
    assert table1.ncols() == 3
    assert table2.is_consistent()
    assert table2.nrows() == 3
    assert table2.ncols() == 1


def test_select():
    # The columns as lists
    c1 = list(range(10))
    c2 = list(range(10))
    c3 = ["a", "b", "c", "d", "e", "f", "g", "i", "j", "k"]

    # Create a table with some elements
    table = flex.reflection_table()
    table["col1"] = flex.int(c1)
    table["col2"] = flex.double(c2)
    table["col3"] = flex.std_string(c3)

    # Select some columns
    new_table = table.select(("col1", "col2"))
    assert new_table.nrows() == 10
    assert new_table.ncols() == 2
    assert all(a == b for a, b in zip(new_table["col1"], c1))
    assert all(a == b for a, b in zip(new_table["col2"], c2))

    # Select some columns
    new_table = table.select(flex.std_string(["col1", "col2"]))
    assert new_table.nrows() == 10
    assert new_table.ncols() == 2
    assert all(a == b for a, b in zip(new_table["col1"], c1))
    assert all(a == b for a, b in zip(new_table["col2"], c2))

    # Select some rows
    index = flex.size_t([0, 1, 5, 8, 9])
    cc1 = [c1[i] for i in index]
    cc2 = [c2[i] for i in index]
    cc3 = [c3[i] for i in index]
    new_table = table.select(index)
    assert new_table.nrows() == 5
    assert new_table.ncols() == 3
    assert all(a == b for a, b in zip(new_table["col1"], cc1))
    assert all(a == b for a, b in zip(new_table["col2"], cc2))
    assert all(a == b for a, b in zip(new_table["col3"], cc3))

    # Select some rows
    index = flex.bool([True, True, False, False, False, True, False, False, True, True])
    new_table = table.select(index)
    assert new_table.nrows() == 5
    assert new_table.ncols() == 3
    assert all(a == b for a, b in zip(new_table["col1"], cc1))
    assert all(a == b for a, b in zip(new_table["col2"], cc2))
    assert all(a == b for a, b in zip(new_table["col3"], cc3))


def test_set_selected():
    # The columns as lists
    c1 = list(range(10))
    c2 = list(range(10))
    c3 = ["a", "b", "c", "d", "e", "f", "g", "i", "j", "k"]

    # Create a table with some elements
    table1 = flex.reflection_table()
    table2 = flex.reflection_table()
    table1["col1"] = flex.int(c1)
    table2["col2"] = flex.double(c2)
    table2["col3"] = flex.std_string(c3)

    # Set selected columns
    table1.set_selected(("col3", "col2"), table2)
    assert table1.nrows() == 10
    assert table1.ncols() == 3
    assert all(a == b for a, b in zip(table1["col1"], c1))
    assert all(a == b for a, b in zip(table1["col2"], c2))
    assert all(a == b for a, b in zip(table1["col3"], c3))

    # Set selected columns
    table1 = flex.reflection_table()
    table1["col1"] = flex.int(c1)
    table1.set_selected(flex.std_string(["col3", "col2"]), table2)
    assert table1.nrows() == 10
    assert table1.ncols() == 3
    assert all(a == b for a, b in zip(table1["col1"], c1))
    assert all(a == b for a, b in zip(table1["col2"], c2))
    assert all(a == b for a, b in zip(table1["col3"], c3))

    cc1 = list(range(10, 15))
    cc2 = list(range(10, 15))
    cc3 = ["l", "m", "n", "o", "p"]

    # Set selected rows
    table2 = flex.reflection_table()
    table2["col1"] = flex.int(cc1)
    table2["col2"] = flex.double(cc2)
    table2["col3"] = flex.std_string(cc3)

    index = flex.size_t([0, 1, 5, 8, 9])
    ccc1 = copy.deepcopy(c1)
    ccc2 = copy.deepcopy(c2)
    ccc3 = copy.deepcopy(c3)
    for j, i in enumerate(index):
        ccc1[i] = cc1[j]
        ccc2[i] = cc2[j]
        ccc3[i] = cc3[j]
    table1.set_selected(index, table2)
    assert all(a == b for a, b in zip(table1["col1"], ccc1))
    assert all(a == b for a, b in zip(table1["col2"], ccc2))
    assert all(a == b for a, b in zip(table1["col3"], ccc3))

    # Set selected rows
    table2 = flex.reflection_table()
    table2["col1"] = flex.int(cc1)
    table2["col2"] = flex.double(cc2)
    table2["col3"] = flex.std_string(cc3)

    table1.set_selected(index, table2)
    assert all(a == b for a, b in zip(table1["col1"], ccc1))
    assert all(a == b for a, b in zip(table1["col2"], ccc2))
    assert all(a == b for a, b in zip(table1["col3"], ccc3))


def test_del_selected():
    # The columns as lists
    c1 = list(range(10))
    c2 = list(range(10))
    c3 = ["a", "b", "c", "d", "e", "f", "g", "i", "j", "k"]

    # Create a table with some elements
    table1 = flex.reflection_table()
    table1["col1"] = flex.int(c1)
    table1["col2"] = flex.double(c2)
    table1["col3"] = flex.std_string(c3)

    # Del selected columns
    table1.del_selected(("col3", "col2"))
    assert table1.nrows() == 10
    assert table1.ncols() == 1
    assert "col1" in table1
    assert "col2" not in table1
    assert "col3" not in table1
    assert all(a == b for a, b in zip(table1["col1"], c1))

    # Del selected columns
    table1 = flex.reflection_table()
    table1["col1"] = flex.int(c1)
    table1["col2"] = flex.double(c2)
    table1["col3"] = flex.std_string(c3)
    table1.del_selected(flex.std_string(["col3", "col2"]))
    assert table1.nrows() == 10
    assert table1.ncols() == 1
    assert "col1" in table1
    assert "col2" not in table1
    assert "col3" not in table1
    assert all(a == b for a, b in zip(table1["col1"], c1))

    # Del selected rows
    table1 = flex.reflection_table()
    table1["col1"] = flex.int(c1)
    table1["col2"] = flex.double(c2)
    table1["col3"] = flex.std_string(c3)

    index = flex.size_t([0, 1, 5, 8, 9])
    index2 = list(range(10))
    for i in index:
        index2.remove(i)
    ccc1 = [c1[i] for i in index2]
    ccc2 = [c2[i] for i in index2]
    ccc3 = [c3[i] for i in index2]
    table1.del_selected(index)
    assert table1.nrows() == len(ccc1)
    assert all(a == b for a, b in zip(table1["col1"], ccc1))
    assert all(a == b for a, b in zip(table1["col2"], ccc2))
    assert all(a == b for a, b in zip(table1["col3"], ccc3))

    # Del selected rows
    table1 = flex.reflection_table()
    table1["col1"] = flex.int(c1)
    table1["col2"] = flex.double(c2)
    table1["col3"] = flex.std_string(c3)

    table1.del_selected(index)
    assert table1.nrows() == len(ccc1)
    assert all(a == b for a, b in zip(table1["col1"], ccc1))
    assert all(a == b for a, b in zip(table1["col2"], ccc2))
    assert all(a == b for a, b in zip(table1["col3"], ccc3))


def test_sort():
    table = flex.reflection_table()
    table["a"] = flex.int([2, 4, 3, 1, 5, 6])
    table["b"] = flex.vec2_double([(3, 2), (3, 1), (1, 3), (4, 5), (4, 3), (2, 0)])
    table["c"] = flex.miller_index(
        [(3, 2, 1), (3, 1, 1), (2, 4, 2), (2, 1, 1), (1, 1, 1), (1, 1, 2)]
    )

    table.sort("a")
    assert list(table["a"]) == [1, 2, 3, 4, 5, 6]

    table.sort("b")
    assert list(table["b"]) == [(1, 3), (2, 0), (3, 1), (3, 2), (4, 3), (4, 5)]

    table.sort("c")
    assert list(table["c"]) == [
        (1, 1, 1),
        (1, 1, 2),
        (2, 1, 1),
        (2, 4, 2),
        (3, 1, 1),
        (3, 2, 1),
    ]

    table.sort("c", order=(1, 2, 0))
    assert list(table["c"]) == [
        (1, 1, 1),
        (2, 1, 1),
        (3, 1, 1),
        (1, 1, 2),
        (3, 2, 1),
        (2, 4, 2),
    ]


def test_flags():
    # Create a table with flags all 0
    table = flex.reflection_table()
    table["flags"] = flex.size_t(5, 0)

    # Get all the flags
    f1 = table.get_flags(table.flags.predicted)
    assert f1.count(True) == 0

    # Set some flags
    mask = flex.bool([True, True, False, False, True])
    table.set_flags(mask, table.flags.predicted)
    f1 = table.get_flags(table.flags.predicted)
    assert f1.count(True) == 3
    assert all(f11 == f22 for f11, f22 in zip(f1, mask))
    f2 = table.get_flags(table.flags.predicted | table.flags.observed)
    assert f2.count(True) == 0

    # Unset the flags
    mask = flex.bool(5, True)
    table.unset_flags(mask, table.flags.predicted | table.flags.observed)
    f1 = table.get_flags(table.flags.predicted)
    assert f1.count(True) == 0
    flags = table["flags"]
    assert all(f == 0 for f in flags)

    # Set multiple flags
    mask = flex.bool([True, True, False, False, True])
    table.set_flags(mask, table.flags.predicted | table.flags.observed)
    f1 = table.get_flags(table.flags.predicted)
    f2 = table.get_flags(table.flags.observed)
    assert f1.count(True) == 3
    assert f2.count(True) == 3
    mask = flex.bool([False, True, True, True, False])
    table.set_flags(mask, table.flags.integrated)
    f1 = table.get_flags(table.flags.predicted)
    f2 = table.get_flags(table.flags.observed)
    f3 = table.get_flags(table.flags.integrated)
    f4 = table.get_flags(table.flags.integrated | table.flags.predicted)
    assert f1.count(True) == 3
    assert f2.count(True) == 3
    assert f3.count(True) == 3
    assert f4.count(True) == 1

    # Get where any are set
    f1 = table.get_flags(table.flags.predicted, all=False)
    f2 = table.get_flags(table.flags.observed, all=False)
    f3 = table.get_flags(table.flags.integrated, all=False)
    f4 = table.get_flags(table.flags.integrated | table.flags.predicted, all=False)
    assert f1.count(True) == 3
    assert f2.count(True) == 3
    assert f3.count(True) == 3
    assert f4.count(True) == 5


def test_serialize():
    # The columns as lists
    c1 = list(range(10))
    c2 = list(range(10))
    c3 = ["a", "b", "c", "d", "e", "f", "g", "i", "j", "k"]

    # Create a table with some elements
    table = flex.reflection_table()
    table["col1"] = flex.int(c1)
    table["col2"] = flex.double(c2)
    table["col3"] = flex.std_string(c3)

    # Pickle, then unpickle
    obj = pickle.dumps(table)
    new_table = pickle.loads(obj)
    assert new_table.is_consistent()
    assert new_table.nrows() == 10
    assert new_table.ncols() == 3
    assert all(a == b for a, b in zip(new_table["col1"], c1))
    assert all(a == b for a, b in zip(new_table["col2"], c2))
    assert all(a == b for a, b in zip(new_table["col3"], c3))


def test_copy():
    # Create a table
    table = flex.reflection_table([("col1", flex.int(range(10)))])

    # Make a shallow copy of the table
    shallow = copy.copy(table)
    shallow["col2"] = flex.double(range(10))
    assert table.ncols() == 2
    assert table.is_consistent()

    # Make a deep copy of the table
    deep = copy.deepcopy(table)
    deep["col3"] = flex.std_string(10)
    assert table.ncols() == 2
    assert deep.ncols() == 3
    assert table.is_consistent()
    assert deep.is_consistent()

    table2 = table.copy()
    table2["col3"] = flex.std_string(10)
    assert table.ncols() == 2
    assert table2.ncols() == 3
    assert table.is_consistent()
    assert table2.is_consistent()


def test_extract_shoeboxes():
    from dials.algorithms.shoebox import MaskCode

    random.seed(0)

    reflections = flex.reflection_table()
    reflections["panel"] = flex.size_t()
    reflections["bbox"] = flex.int6()

    npanels = 2
    width = 1000
    height = 1000
    frame0 = 10
    frame1 = 100
    nrefl = 1000

    for i in range(nrefl):
        xs = random.randint(5, 10)
        ys = random.randint(5, 10)
        x0 = random.randint(-xs + 1, width - 1)
        y0 = random.randint(-ys + 1, height - 1)
        z0 = random.randint(frame0, frame1 - 1)
        x1 = x0 + xs
        y1 = y0 + ys
        z1 = min([z0 + random.randint(1, 10), frame1])
        assert x1 > x0
        assert y1 > y0
        assert z1 > z0
        assert z0 >= frame0 and z1 <= frame1
        bbox = (x0, x1, y0, y1, z0, z1)
        reflections.append({"panel": random.randint(0, 1), "bbox": bbox})

    reflections["shoebox"] = flex.shoebox(reflections["panel"], reflections["bbox"])
    reflections["shoebox"].allocate()

    class FakeImageSet:
        def __init__(self):
            self.data = flex.int(range(height * width))
            self.data.reshape(flex.grid(height, width))

        def get_array_range(self):
            return (frame0, frame1)

        def get_detector(self):
            class FakeDetector:
                def __len__(self):
                    return npanels

                def __getitem__(self, index):
                    class FakePanel:
                        def get_trusted_range(self):
                            return (0, 1000000)

                    return FakePanel()

            return FakeDetector()

        def __len__(self):
            return frame1 - frame0

        def __getitem__(self, index):
            f = frame0 + index
            return (self.data + f * 1, self.data + f * 2)

        def get_corrected_data(self, index):
            f = frame0 + index
            return (self.data + f * 1, self.data + f * 2)

        def get_mask(self, index):
            image = self.get_corrected_data(index)
            return tuple(im >= 0 for im in image)

    imageset = FakeImageSet()

    reflections.extract_shoeboxes(imageset)

    for i in range(len(reflections)):
        sbox = reflections[i]["shoebox"]
        assert sbox.is_consistent()
        mask = sbox.mask
        data = sbox.data
        bbox = sbox.bbox
        panel = sbox.panel
        x0, x1, y0, y1, z0, z1 = bbox
        for z in range(z1 - z0):
            for y in range(y1 - y0):
                for x in range(x1 - x0):
                    v1 = data[z, y, x]
                    m1 = mask[z, y, x]
                    if (
                        x0 + x >= 0
                        and y0 + y >= 0
                        and x0 + x < width
                        and y0 + y < height
                    ):
                        v2 = imageset.data[y + y0, x + x0] + (z + z0) * (panel + 1)
                        m2 = MaskCode.Valid
                        assert v1 == v2
                        assert m1 == m2
                    else:
                        assert v1 == 0
                        assert m1 == 0


def test_split_by_experiment_id():
    r = flex.reflection_table()
    r["id"] = flex.int()
    for i in range(100):
        r.append({"id": 0})
        r.append({"id": 1})
        r.append({"id": 2})
        r.append({"id": 3})
        r.append({"id": 5})
    result = r.split_by_experiment_id()
    assert len(result) == 5
    for res, exp in zip(result, [0, 1, 2, 3, 5]):
        assert len(res) == 100
        assert res["id"].count(exp) == 100

    # test the same but with experiment_identifiers() set - keep separate as
    # function must work with and without experiment_identifiers() set
    r.experiment_identifiers()[0] = "0"
    r.experiment_identifiers()[1] = "1"
    r.experiment_identifiers()[2] = "2"
    r.experiment_identifiers()[3] = "3"
    r.experiment_identifiers()[5] = "5"
    result = r.split_by_experiment_id()
    assert len(result) == 5
    for res, exp in zip(result, [0, 1, 2, 3, 5]):
        assert len(res) == 100
        assert res["id"].count(exp) == 100
        assert list(res.experiment_identifiers().keys()) == [exp]
        assert list(res.experiment_identifiers().values()) == [str(exp)]


def test_split_indices_by_experiment_id():
    r = flex.reflection_table()
    r["id"] = flex.int()
    for i in range(100):
        r.append({"id": 0})
        r.append({"id": 1})
        r.append({"id": 2})
        r.append({"id": 3})
        r.append({"id": 5})
    index_list = r.split_indices_by_experiment_id(6)
    assert len(index_list) == 6
    for index, exp, num in zip(
        index_list, [0, 1, 2, 3, 4, 5], [100, 100, 100, 100, 0, 100]
    ):
        assert len(index) == num
        assert r.select(index)["id"].count(exp) == num


def test_split_partials():
    r = flex.reflection_table()
    r["value1"] = flex.double()
    r["value2"] = flex.int()
    r["value3"] = flex.double()
    r["bbox"] = flex.int6()
    expected = []
    for i in range(100):
        x0 = random.randint(0, 100)
        x1 = x0 + random.randint(1, 10)
        y0 = random.randint(0, 100)
        y1 = y0 + random.randint(1, 10)
        z0 = random.randint(0, 100)
        z1 = z0 + random.randint(1, 10)
        v1 = random.uniform(0, 100)
        v2 = random.randint(0, 100)
        v3 = random.uniform(0, 100)
        r.append(
            {"value1": v1, "value2": v2, "value3": v3, "bbox": (x0, x1, y0, y1, z0, z1)}
        )
        for z in range(z0, z1):
            expected.append(
                {
                    "value1": v1,
                    "value2": v2,
                    "value3": v3,
                    "bbox": (x0, x1, y0, y1, z, z + 1),
                    "partial_id": i,
                }
            )

    r.split_partials()
    assert len(r) == len(expected)
    EPS = 1e-7
    for r1, r2 in zip(r.rows(), expected):
        assert abs(r1["value1"] - r2["value1"]) < EPS
        assert r1["value2"] == r2["value2"]
        assert abs(r1["value3"] - r2["value3"]) < EPS
        assert r1["bbox"] == r2["bbox"]
        assert r1["partial_id"] == r2["partial_id"]


def test_split_partials_with_shoebox():
    from dials.model.data import Shoebox

    r = flex.reflection_table()
    r["value1"] = flex.double()
    r["value2"] = flex.int()
    r["value3"] = flex.double()
    r["bbox"] = flex.int6()
    r["panel"] = flex.size_t()
    r["shoebox"] = flex.shoebox()
    expected = []
    for i in range(100):
        x0 = random.randint(0, 100)
        x1 = x0 + random.randint(1, 10)
        y0 = random.randint(0, 100)
        y1 = y0 + random.randint(1, 10)
        z0 = random.randint(0, 100)
        z1 = z0 + random.randint(1, 10)
        v1 = random.uniform(0, 100)
        v2 = random.randint(0, 100)
        v3 = random.uniform(0, 100)
        sbox = Shoebox(0, (x0, x1, y0, y1, z0, z1))
        sbox.allocate()
        assert sbox.is_consistent()
        w = x1 - x0
        h = y1 - y0
        for z in range(z0, z1):
            for y in range(y0, y1):
                for x in range(x0, x1):
                    sbox.data[z - z0, y - y0, x - x0] = x + y * w + z * w * h
        r.append(
            {
                "value1": v1,
                "value2": v2,
                "value3": v3,
                "bbox": (x0, x1, y0, y1, z0, z1),
                "panel": 0,
                "shoebox": sbox,
            }
        )
        for z in range(z0, z1):
            sbox = Shoebox(0, (x0, x1, y0, y1, z, z + 1))
            sbox.allocate()
            assert sbox.is_consistent()
            w = x1 - x0
            h = y1 - y0
            for y in range(y0, y1):
                for x in range(x0, x1):
                    sbox.data[0, y - y0, x - x0] = x + y * w + z * w * h
            expected.append(
                {
                    "value1": v1,
                    "value2": v2,
                    "value3": v3,
                    "bbox": (x0, x1, y0, y1, z, z + 1),
                    "partial_id": i,
                    "panel": 0,
                    "shoebox": sbox,
                }
            )

    r.split_partials_with_shoebox()
    assert len(r) == len(expected)
    EPS = 1e-7
    for r1, r2 in zip(r.rows(), expected):
        assert abs(r1["value1"] - r2["value1"]) < EPS
        assert r1["value2"] == r2["value2"]
        assert abs(r1["value3"] - r2["value3"]) < EPS
        assert r1["bbox"] == r2["bbox"]
        assert r1["partial_id"] == r2["partial_id"]
        assert r1["panel"] == r2["panel"]
        assert (
            r1["shoebox"]
            .data.as_double()
            .as_1d()
            .all_approx_equal(r2["shoebox"].data.as_double().as_1d())
        )


def test_find_overlapping():
    N = 10000
    r = flex.reflection_table(N)
    r["bbox"] = flex.int6(N)
    r["panel"] = flex.size_t(N)
    r["id"] = flex.int(N)
    r["imageset_id"] = flex.int(N)
    for i in range(N):
        x0 = random.randint(0, 100)
        x1 = random.randint(1, 10) + x0
        y0 = random.randint(0, 100)
        y1 = random.randint(1, 10) + y0
        z0 = random.randint(0, 100)
        z1 = random.randint(1, 10) + z0
        panel = random.randint(0, 2)
        pid = random.randint(0, 2)
        r["bbox"][i] = (x0, x1, y0, y1, z0, z1)
        r["panel"][i] = panel
        r["id"][i] = pid
        r["imageset_id"][i] = pid

    def is_overlap(b0, b1, border):
        b0 = (
            b0[0] - border,
            b0[1] + border,
            b0[2] - border,
            b0[3] + border,
            b0[4] - border,
            b0[5] + border,
        )
        b1 = (
            b1[0] - border,
            b1[1] + border,
            b1[2] - border,
            b1[3] + border,
            b1[4] - border,
            b1[5] + border,
        )
        if not (
            b1[0] > b0[1]
            or b1[1] < b0[0]
            or b1[2] > b0[3]
            or b1[3] < b0[2]
            or b1[4] > b0[5]
            or b1[5] < b0[4]
        ):
            return True
        return False

    for i in [0, 2, 5]:
        overlaps = r.find_overlaps(border=i)
        for item in overlaps.edges():
            i0 = overlaps.source(item)
            i1 = overlaps.target(item)
            r0 = r[i0]
            r1 = r[i1]
            p0 = r0["panel"]
            p1 = r1["panel"]
            b0 = r0["bbox"]
            b1 = r1["bbox"]
            j0 = r0["imageset_id"]
            j1 = r1["imageset_id"]
            assert j0 == j1
            assert p0 == p1
            assert is_overlap(b0, b1, i)


def test_to_from_msgpack(tmpdir):
    from dials.model.data import Shoebox

    def gen_shoebox():
        shoebox = Shoebox(0, (0, 4, 0, 3, 0, 1))
        shoebox.allocate()
        for k in range(1):
            for j in range(3):
                for i in range(4):
                    shoebox.data[k, j, i] = i + j + k + 0.1
                    shoebox.mask[k, j, i] = i % 2
                    shoebox.background[k, j, i] = i * j + 0.2
        return shoebox

    def compare(a, b):
        assert a.is_consistent()
        assert b.is_consistent()
        assert a.panel == b.panel
        assert a.bbox == b.bbox
        for aa, bb in zip(a.data, b.data):
            if abs(aa - bb) > 1e-9:
                return False
        for aa, bb in zip(a.background, b.background):
            if abs(aa - bb) > 1e-9:
                return False
        for aa, bb in zip(a.mask, b.mask):
            if aa != bb:
                return False
        return True

    # The columns as lists
    c1 = list(range(10))
    c2 = list(range(10))
    c3 = ["a", "b", "c", "d", "e", "f", "g", "i", "j", "k"]
    c4 = [True, False, True, False, True] * 2
    c5 = list(range(10))
    c6 = [(i + 1, i + 2) for i in range(10)]
    c7 = [(i + 1, i + 2, i + 3) for i in range(10)]
    c8 = [tuple(i + j for j in range(9)) for i in range(10)]
    c9 = [tuple(i + j for j in range(6)) for i in range(10)]
    c10 = [(i + 1, i + 2, i + 3) for i in range(10)]
    c11 = [gen_shoebox() for i in range(10)]

    # Create a table with some elements
    table = flex.reflection_table()
    table["col1"] = flex.int(c1)
    table["col2"] = flex.double(c2)
    table["col3"] = flex.std_string(c3)
    table["col4"] = flex.bool(c4)
    table["col5"] = flex.size_t(c5)
    table["col6"] = flex.vec2_double(c6)
    table["col7"] = flex.vec3_double(c7)
    table["col8"] = flex.mat3_double(c8)
    table["col9"] = flex.int6(c9)
    table["col10"] = flex.miller_index(c10)
    table["col11"] = flex.shoebox(c11)

    obj = table.as_msgpack()
    new_table = flex.reflection_table.from_msgpack(obj)
    assert new_table.is_consistent()
    assert new_table.nrows() == 10
    assert new_table.ncols() == 11
    assert all(tuple(a == b for a, b in zip(new_table["col1"], c1)))
    assert all(tuple(a == b for a, b in zip(new_table["col2"], c2)))
    assert all(tuple(a == b for a, b in zip(new_table["col3"], c3)))
    assert all(tuple(a == b for a, b in zip(new_table["col4"], c4)))
    assert all(tuple(a == b for a, b in zip(new_table["col5"], c5)))
    assert all(tuple(a == b for a, b in zip(new_table["col6"], c6)))
    assert all(tuple(a == b for a, b in zip(new_table["col7"], c7)))
    assert all(tuple(a == b for a, b in zip(new_table["col8"], c8)))
    assert all(tuple(a == b for a, b in zip(new_table["col9"], c9)))
    assert all(tuple(a == b for a, b in zip(new_table["col10"], c10)))
    assert all(tuple(compare(a, b) for a, b in zip(new_table["col11"], c11)))

    table.as_msgpack_file(tmpdir.join("reflections.mpack").strpath)
    new_table = flex.reflection_table.from_msgpack_file(
        tmpdir.join("reflections.mpack").strpath
    )
    assert new_table.is_consistent()
    assert new_table.nrows() == 10
    assert new_table.ncols() == 11
    assert all(tuple(a == b for a, b in zip(new_table["col1"], c1)))
    assert all(tuple(a == b for a, b in zip(new_table["col2"], c2)))
    assert all(tuple(a == b for a, b in zip(new_table["col3"], c3)))
    assert all(tuple(a == b for a, b in zip(new_table["col4"], c4)))
    assert all(tuple(a == b for a, b in zip(new_table["col5"], c5)))
    assert all(tuple(a == b for a, b in zip(new_table["col6"], c6)))
    assert all(tuple(a == b for a, b in zip(new_table["col7"], c7)))
    assert all(tuple(a == b for a, b in zip(new_table["col8"], c8)))
    assert all(tuple(a == b for a, b in zip(new_table["col9"], c9)))
    assert all(tuple(a == b for a, b in zip(new_table["col10"], c10)))
    assert all(tuple(compare(a, b) for a, b in zip(new_table["col11"], c11)))


def test_experiment_identifiers():
    from dxtbx.model import Experiment, ExperimentList

    table = flex.reflection_table()
    table["id"] = flex.int([0, 1, 2, 3])

    table.assert_experiment_identifiers_are_consistent()

    identifiers = table.experiment_identifiers()
    identifiers[0] = "abcd"
    identifiers[1] = "efgh"
    identifiers[2] = "ijkl"
    identifiers[3] = "mnop"

    assert identifiers[0] == "abcd"
    assert identifiers[1] == "efgh"
    assert identifiers[2] == "ijkl"
    assert identifiers[3] == "mnop"

    for k, v in identifiers:
        if k == 0:
            assert v == "abcd"
        if k == 1:
            assert v == "efgh"
        if k == 2:
            assert v == "ijkl"
        if k == 3:
            assert v == "mnop"

    assert tuple(identifiers.keys()) == (0, 1, 2, 3)
    assert tuple(identifiers.values()) == ("abcd", "efgh", "ijkl", "mnop")

    table.assert_experiment_identifiers_are_consistent()

    experiments = ExperimentList()
    experiments.append(Experiment(identifier="abcd"))
    experiments.append(Experiment(identifier="efgh"))
    experiments.append(Experiment(identifier="ijkl"))
    experiments.append(Experiment(identifier="mnop"))

    table.assert_experiment_identifiers_are_consistent()

    experiments = ExperimentList()
    experiments.append(Experiment(identifier="abcd"))
    experiments.append(Experiment(identifier="efgh"))
    experiments.append(Experiment(identifier="ijkl"))
    experiments.append(Experiment(identifier="mnop"))
    experiments[3].identifier = "ijkl"

    with pytest.raises(AssertionError):
        table.assert_experiment_identifiers_are_consistent(experiments)

    experiments[2].identifier = "mnop"
    table.assert_experiment_identifiers_are_consistent(experiments)

    identifiers = table.experiment_identifiers()
    identifiers[0] = "abcd"
    identifiers[1] = "efgh"
    identifiers[2] = "ijkl"
    identifiers[3] = "ijkl"

    with pytest.raises(AssertionError):
        table.assert_experiment_identifiers_are_consistent()

    identifiers[3] = "mnop"

    pickled = pickle.dumps(table)
    table2 = pickle.loads(pickled)

    id1 = table.experiment_identifiers()
    id2 = table2.experiment_identifiers()

    for i in id1.keys():
        assert id1[i] == id2[i]

    other_table = flex.reflection_table()
    other_table["id"] = flex.int([3, 4])

    table.assert_experiment_identifiers_are_consistent()

    packed = table.as_msgpack()
    table2 = table.from_msgpack(packed)

    id1 = table.experiment_identifiers()
    id2 = table2.experiment_identifiers()

    for i in id1.keys():
        assert id1[i] == id2[i]

    other_table = flex.reflection_table()
    other_table["id"] = flex.int([3, 4])

    table.assert_experiment_identifiers_are_consistent()

    identifiers = other_table.experiment_identifiers()
    identifiers[3] = "mnop"
    identifiers[4] = "qrst"

    table.extend(other_table)

    assert len(table.experiment_identifiers()) == 5
    assert table.experiment_identifiers()[0] == "abcd"
    assert table.experiment_identifiers()[1] == "efgh"
    assert table.experiment_identifiers()[2] == "ijkl"
    assert table.experiment_identifiers()[3] == "mnop"
    assert table.experiment_identifiers()[4] == "qrst"

    assert len(table.experiment_identifiers()) == 5
    assert table.experiment_identifiers()[0] == "abcd"
    assert table.experiment_identifiers()[1] == "efgh"
    assert table.experiment_identifiers()[2] == "ijkl"
    assert table.experiment_identifiers()[3] == "mnop"
    assert table.experiment_identifiers()[4] == "qrst"


def test_select_remove_on_experiment_identifiers(caplog):

    table = flex.reflection_table()
    table["id"] = flex.int([0, 1, 2, 3])

    experiments = ExperimentList()
    experiments.append(Experiment(identifier="abcd"))
    experiments.append(Experiment(identifier="efgh"))
    experiments.append(Experiment(identifier="ijkl"))
    experiments.append(Experiment(identifier="mnop"))
    table.experiment_identifiers()[0] = "abcd"
    table.experiment_identifiers()[1] = "efgh"
    table.experiment_identifiers()[2] = "ijkl"
    table.experiment_identifiers()[3] = "mnop"

    table.assert_experiment_identifiers_are_consistent(experiments)

    table = table.remove_on_experiment_identifiers(["efgh"])
    del experiments[1]
    table.assert_experiment_identifiers_are_consistent(experiments)

    assert list(table.experiment_identifiers().keys()) == [0, 2, 3]
    assert list(table.experiment_identifiers().values()) == ["abcd", "ijkl", "mnop"]

    table = table.select_on_experiment_identifiers(["abcd", "mnop"])
    del experiments[1]  # now ijkl
    table.assert_experiment_identifiers_are_consistent(experiments)
    assert list(table.experiment_identifiers().keys()) == [0, 3]
    assert list(table.experiment_identifiers().values()) == ["abcd", "mnop"]

    # reset 'id' column such that they are numbered 0 .. n-1
    table.reset_ids()
    table.assert_experiment_identifiers_are_consistent(experiments)
    assert list(table.experiment_identifiers().keys()) == [0, 1]
    assert list(table.experiment_identifiers().values()) == ["abcd", "mnop"]
    # test that the function doesn't fail if no identifiers set
    table1 = copy.deepcopy(table)
    for k in table1.experiment_identifiers().keys():
        del table1.experiment_identifiers()[k]
    table1.reset_ids()
    assert list(table1.experiment_identifiers().keys()) == []

    # Test warning is logged if bad choice
    caplog.set_level(logging.WARNING)
    table.remove_on_experiment_identifiers(["efgh"])
    assert caplog.record_tuples[0][1] == logging.WARNING
    caplog.clear()

    table.select_on_experiment_identifiers(["efgh"])
    assert caplog.record_tuples[0][1] == logging.WARNING
    caplog.clear()

    table = flex.reflection_table()
    table["id"] = flex.int([0, 1, 2, 3])
    # Test warning is logged if identifiers map not set

    table.remove_on_experiment_identifiers(["efgh"])
    assert caplog.record_tuples[0][1] == logging.WARNING
    caplog.clear()

    table.select_on_experiment_identifiers(["abcd", "mnop"])


def test_as_miller_array():
    table = flex.reflection_table()
    table["intensity.1.value"] = flex.double([1.0, 2.0, 3.0])
    table["intensity.1.variance"] = flex.double([0.25, 1.0, 4.0])
    table["miller_index"] = flex.miller_index([(1, 0, 0), (2, 0, 0), (3, 0, 0)])

    crystal = Crystal(
        real_space_a=(10, 0, 0),
        real_space_b=(0, 11, 0),
        real_space_c=(0, 0, 12),
        space_group=sgtbx.space_group_info("P 222").group(),
    )
    experiment = Experiment(crystal=crystal)

    iobs = table.as_miller_array(experiment, intensity="1")
    assert list(iobs.data()) == list(table["intensity.1.value"])
    assert list(iobs.sigmas()) == list(flex.sqrt(table["intensity.1.variance"]))

    with pytest.raises(KeyError):
        _ = table.as_miller_array(experiment, intensity="2")
    table["intensity.2.value"] = flex.double([1.0, 2.0, 3.0])
    with pytest.raises(KeyError):
        _ = table.as_miller_array(experiment, intensity="2")


def test_map_centroids_to_reciprocal_space(dials_regression):
    data_dir = os.path.join(dials_regression, "indexing_test_data", "i04_weak_data")
    pickle_path = os.path.join(data_dir, "full.pickle")
    expts_path = os.path.join(data_dir, "experiments_import.json")

    refl = flex.reflection_table.from_file(pickle_path)
    expts = load.experiment_list(expts_path, check_format=False)

    # check mm values not in
    assert "xyzobs.mm.value" not in refl

    refl.centroid_px_to_mm(expts)

    for k in ("xyzobs.mm.value", "xyzobs.mm.variance"):
        assert k in refl

    assert refl["xyzobs.mm.value"][0] == pytest.approx(
        (199.43400000000003, 11.908133333333334, 1.4324789835743459)
    )
    assert refl["xyzobs.mm.variance"][0] == pytest.approx(
        (0.0035346345381526106, 0.0029881028112449803, 5.711576621000785e-07)
    )

    refl.map_centroids_to_reciprocal_space(expts)

    for k in ("s1", "rlp"):
        assert k in refl

    assert refl["s1"][0] == pytest.approx(
        (-0.035321308540942425, 0.6030297672949761, -0.8272574664632307)
    )
    assert refl["rlp"][0] == pytest.approx(
        (-0.035321308540942425, 0.27833194706770875, -0.5700990597173606)
    )

    # select only those centroids on the first image
    sel = refl["xyzobs.px.value"].parts()[2] < 1
    refl1 = refl.select(sel)
    del refl1["xyzobs.mm.value"], refl1["xyzobs.mm.variance"], refl1["s1"], refl1["rlp"]

    # pretend this is a still and hence no scan or goniometer
    expts[0].goniometer = None
    expts[0].scan = None
    refl1.centroid_px_to_mm(expts)
    refl1.map_centroids_to_reciprocal_space(expts)

    assert refl1["s1"][0] == pytest.approx(
        (-0.035321308540942425, 0.6030297672949761, -0.8272574664632307)
    )
    # numbers for rlp are different to above since for the goniometer case the
    # starting angle of the first image is non-zero, so the rlps are rotated back
    # to zero degrees
    assert refl1["rlp"][0] == pytest.approx(
        (-0.035321308540942425, 0.6030297672949761, 0.19707031842793443)
    )


def test_calculate_entering_flags(dials_regression):
    data_dir = os.path.join(dials_regression, "indexing_test_data", "i04_weak_data")
    pickle_path = os.path.join(data_dir, "full.pickle")
    experiments_path = os.path.join(data_dir, "experiments_import.json")

    refl = flex.reflection_table.from_pickle(pickle_path)
    experiments = load.experiment_list(experiments_path, check_format=False)

    refl.centroid_px_to_mm(experiments)
    refl.map_centroids_to_reciprocal_space(experiments)
    refl.calculate_entering_flags(experiments)
    assert "entering" in refl
    flags = refl["entering"]
    assert flags.count(True) == 58283
    assert flags.count(False) == 57799


def test_random_split():
    """Test the reflection_table.random_split() method.

    A random seed is set, so the result is reproducible.
    """
    flex.set_random_seed(0)
    table = flex.reflection_table()
    table["id"] = flex.int(range(0, 10))

    # first test splitting into 2 or 3 tables.
    split_tables = table.random_split(2)
    expected = [[5, 7, 3, 0, 8], [2, 9, 6, 1, 4]]
    for t, e in zip(split_tables, expected):
        assert list(t["id"]) == e

    split_tables = table.random_split(3)
    expected = [[5, 7, 0], [4, 3, 8], [1, 9, 2, 6]]
    for t, e in zip(split_tables, expected):
        assert list(t["id"]) == e

    # test that a float is handled
    split_tables = table.random_split(1.0)
    assert split_tables[0] is table

    # values below 1 should just return the table
    split_tables = table.random_split(0)
    assert split_tables[0] is table

    # if n > len(table), the table should split into n=len(table) tables with
    # one entry.
    split_tables = table.random_split(20)
    assert len(split_tables) == 10
    for t in split_tables:
        assert len(t) == 1


def test_match_basic():
    n = 100
    s = 10

    def r():
        return random.random()

    xyz = flex.vec3_double()

    for j in range(n):
        xyz.append((r() * s, r() * s, r() * s * 20))

    order = list(range(n))

    random.shuffle(order)

    xyz2 = xyz.select(flex.size_t(order))

    a = flex.reflection_table()
    b = flex.reflection_table()

    a["xyz"] = xyz
    b["xyz"] = xyz2

    mm, nn, distance = a.match(b, key="xyz", scale=(1.0, 1.0, 0.05))

    a_ = a.select(mm)
    b_ = b.select(nn)

    for _a, _b in zip(a_["xyz"], b_["xyz"]):
        assert _a == pytest.approx(_b)


def test_match_mismatched_sizes():
    n = 100
    s = 10

    def r():
        return random.random()

    xyz = flex.vec3_double()

    for j in range(n):
        xyz.append((r() * s, r() * s, r() * s * 20))

    order = list(range(n))

    random.shuffle(order)

    xyz2 = xyz.select(flex.size_t(order))

    a = flex.reflection_table()
    b = flex.reflection_table()

    a["xyz"] = xyz[: n // 2]
    b["xyz"] = xyz2

    mm, nn, distance = a.match(b, key="xyz", scale=(1.0, 1.0, 0.05))

    a_ = a.select(mm)
    b_ = b.select(nn)

    for _a, _b in zip(a_["xyz"], b_["xyz"]):
        assert _a == pytest.approx(_b)


def test_match_by_hkle():
    nn = 10

    h = flex.int([n % nn for n in range(nn)])
    k = flex.int([(n + 2) % nn for n in range(nn)])
    l = flex.int([(n + 4) % nn for n in range(nn)])
    e = flex.int([n % 2 for n in range(nn)])

    hkl = flex.miller_index(h, k, l)

    t0 = flex.reflection_table()
    t0["miller_index"] = hkl
    t0["entering"] = e

    i = list(range(nn))
    random.shuffle(i)
    t1 = t0.select(flex.size_t(i))

    # because t0.match_by_hkle(t1) will give the _inverse_ to i
    n0, n1 = t1.match_by_hkle(t0)

    assert list(n0) == list(range(nn))
    assert list(n1) == i


def test_concat():

    table1 = flex.reflection_table()
    table2 = flex.reflection_table()

    table1["id"] = flex.size_t([0, 0, 1, 1])
    table2["id"] = flex.size_t([0, 0, 1, 1])

    ids1 = table1.experiment_identifiers()
    ids2 = table2.experiment_identifiers()

    ids1[0] = "a"
    ids1[1] = "b"
    ids2[0] = "c"
    ids2[1] = "d"

    table1 = flex.reflection_table.concat([table1, table2])

    assert list(table1["id"]) == [0, 0, 1, 1, 2, 2, 3, 3]
    assert list(ids1.keys()) == [0, 1, 2, 3]
    assert list(ids1.values()) == ["a", "b", "c", "d"]



from __future__ import annotations

from dials_algorithms_image_filter_ext import *  # noqa: F403; lgtm

__all__ = (  # noqa: F405
    "IndexOfDispersionFilterDouble",
    "IndexOfDispersionFilterFloat",
    "IndexOfDispersionFilterMaskedDouble",
    "IndexOfDispersionFilterMaskedFloat",
    "MeanAndVarianceFilterDouble",
    "MeanAndVarianceFilterFloat",
    "MeanAndVarianceFilterMaskedDouble",
    "MeanAndVarianceFilterMaskedFloat",
    "anisotropic_diffusion",
    "chebyshev_distance",
    "convolve",
    "convolve_col",
    "convolve_row",
    "index_of_dispersion_filter",
    "manhattan_distance",
    "mean_and_variance_filter",
    "mean_filter",
    "median_filter",
    "summed_area",
    "summed_area_table",
)



from __future__ import annotations

import random
from unittest import mock

from cctbx import sgtbx
from dxtbx.model import Crystal, Experiment, ExperimentList
from scitbx import matrix

from dials.algorithms.clustering import observers
from dials.algorithms.clustering.unit_cell import cluster_unit_cells


def test_UnitCellAnalysisObserver():
    # generate some random unit cells
    sgi = sgtbx.space_group_info("P1")
    unit_cells = [
        sgi.any_compatible_unit_cell(volume=random.uniform(990, 1010))
        for i in range(10)
    ]

    # generate experiment list
    experiments = ExperimentList()
    U = matrix.identity(3)
    for uc in unit_cells:
        B = matrix.sqr(uc.fractionalization_matrix()).transpose()
        direct_matrix = (U * B).inverse()
        experiments.append(
            Experiment(
                crystal=Crystal(
                    direct_matrix[:3],
                    direct_matrix[3:6],
                    direct_matrix[6:9],
                    space_group=sgi.group(),
                )
            )
        )

    # generate dendrogram
    crystal_symmetries = [expt.crystal.get_crystal_symmetry() for expt in experiments]
    result = cluster_unit_cells(
        crystal_symmetries, lattice_ids=list(experiments.identifiers())
    )

    # setup script
    script = mock.Mock()
    script._experiments = experiments
    script.unit_cell_dendrogram = result.dendrogram

    # test the observer
    observer = observers.UnitCellAnalysisObserver()
    observer.update(script)
    assert set(observer.data) == {"experiments", "dendrogram"}
    d = observer.make_plots()
    assert "unit_cell_graphs" in d



from __future__ import annotations

import procrunner


def test_export_best(dials_data, tmp_path):
    result = procrunner.run(
        [
            "dials.import",
            "template="
            + str(dials_data("centroid_test_data", pathlib=True) / "centroid_####.cbf"),
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    result = procrunner.run(
        ["dials.find_spots", "imported.expt", "nproc=1"], working_directory=tmp_path
    )
    assert not result.returncode and not result.stderr
    result = procrunner.run(
        ["dials.index", "imported.expt", "strong.refl", "space_group=P422"],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    result = procrunner.run(
        [
            "dials.integrate",
            "nproc=1",
            "indexed.expt",
            "indexed.refl",
            "prediction.padding=0",
            "sigma_m_algorithm=basic",
        ],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr
    result = procrunner.run(
        ["dials.export_best", "integrated.expt", "integrated.refl"],
        working_directory=tmp_path,
    )
    assert not result.returncode and not result.stderr

    assert tmp_path.joinpath("best.dat").is_file()
    assert tmp_path.joinpath("best.hkl").is_file()
    assert tmp_path.joinpath("best.par").is_file()

    with tmp_path.joinpath("best.dat").open("r") as f:
        lines = "".join(f.readlines()[:10])
    assert (
        lines
        == """\
  181.8877       0.77       1.60
   63.1895       1.59       1.81
   38.2372       1.87       1.71
   27.4131       1.84       1.55
   21.3655       1.89       1.51
   17.5043       1.88       1.49
   14.8254       1.89       1.45
   12.8580       1.91       1.45
   11.3518       1.89       1.42
   10.1617       1.87       1.41
"""
    )

    with tmp_path.joinpath("best.hkl").open("r") as f:
        lines = "".join(f.readlines()[:10])
    assert (
        lines
        == """\
 -20   27   -8      20.17      20.00
 -20   27   -7      74.13      21.59
 -20   27   -6      22.34      19.57
 -20   27   -5       6.33      19.72
 -20   28  -10      19.77      18.77
 -20   28   -9      50.37      20.28
 -20   28   -7      69.23      21.42
 -20   28   -6      24.42      19.56
 -20   28   -4     -10.35      19.51
 -20   28   -2      47.53      20.49
"""
    )

    lines = tmp_path.joinpath("best.par").read_text()
    assert (
        lines
        == """\
# parameter file for BEST
TITLE          From DIALS
DETECTOR       PILA
SITE           Not set
DIAMETER       434.64
PIXEL          0.172
ROTAXIS        0.01 0.00 1.00 FAST
POLAXIS        0.00 1.00 0.00
GAIN               1.00
CMOSAIC            0.54
PHISTART           0.00
PHIWIDTH           0.20
DISTANCE         191.09
WAVELENGTH      0.97950
POLARISATION    0.99900
SYMMETRY       P422
UB             -0.012248 -0.020067  0.003152
               -0.005029 -0.000351 -0.024623
                0.019651 -0.012597 -0.004336
CELL              42.20    42.20    39.68  90.00  90.00  90.00
RASTER           7 7 5 3 3
SEPARATION      0.665  0.665
BEAM            219.875  212.612
# end of parameter file for BEST
"""
    )



# XXX This is not a standalone installer!  It must be used as part of the
# framework in libtbx/auto_build.
"""
Installer script for DIALS, based on automatically generated template.  This
must be moved to the proper location to work.
"""


from __future__ import annotations

import os.path
import sys

from libtbx.auto_build import install_distribution

libtbx_path = os.path.join(
    os.path.abspath(os.path.dirname(os.path.dirname(__file__))), "lib"
)
if libtbx_path not in sys.path:
    sys.path.append(libtbx_path)


class installer(install_distribution.installer):
    # XXX most settings can be edited here
    product_name = "DIALS"
    dest_dir_prefix = "dials"
    make_apps = []
    configure_modules = install_distribution.installer.configure_modules + [
        "dials",
        "cbflib",
        "annlib_adaptbx",
        "wxtbx",
        "gltbx",
    ]
    include_gui_packages = True
    remove_sources_default = False  # XXX toggles removal of C++ files after building
    base_package_options = ["--dials", "--all"]
    source_packages = ["cctbx_bundle"] + ["dials", "cbflib", "annlib", "annlib_adaptbx"]
    #

    installer_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

    # Various functions specific to DIALS go below here
    #
    def product_specific_finalize_install(self, log):
        # XXX remove 'phenix' dispatchers from bin directory
        bin_dir = os.path.join(self.build_dir, "bin")
        for file_name in os.listdir(bin_dir):
            if file_name.startswith("phenix"):
                os.remove(os.path.join(bin_dir, file_name))


if __name__ == "__main__":
    installer(sys.argv[1:])



from __future__ import annotations

import sys

from libtbx import phil

from dials.array_family import flex
from dials.util import Sorry, show_mail_handle_errors
from dials.util.multi_dataset_handling import (
    assign_unique_identifiers,
    parse_multiple_datasets,
)
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files

help_message = """Command line script which assigns experiment identifiers
to reflections and experiments and saves them back to disk.
"""

phil_scope = phil.parse(
    """
  identifiers = None
    .type = strings
    .help = "User specified identifiers to use, must be a list of strings equal"
            "to the number of datasets."
  output {
    reflections = assigned.refl
      .type = str
    experiments = assigned.expt
      .type = str
  }
"""
)


@show_mail_handle_errors()
def run(args=None):
    """Run assign experiment identifiers from the command line."""
    usage = """
    dials.assign_experiment_identifiers observations.refl models.expt

    dials.assign_experiment_identifiers observations.refl models.expt identifiers="0 1 2"
"""
    parser = ArgumentParser(
        usage=usage,
        read_experiments=True,
        read_reflections=True,
        phil=phil_scope,
        check_format=False,
        epilog=help_message,
    )
    params, _ = parser.parse_args(args=args, show_diff_phil=False)

    if not params.input.experiments or not params.input.reflections:
        parser.print_help()
        sys.exit()

    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    reflections = parse_multiple_datasets(reflections)
    if len(experiments) != len(reflections):
        raise Sorry(
            "Mismatched number of experiments and reflection tables found: %s & %s."
            % (len(experiments), len(reflections))
        )
    try:
        experiments, reflections = assign_unique_identifiers(
            experiments, reflections, params.identifiers
        )
    except ValueError as e:
        raise Sorry(e)
    print(f"assigned identifiers: {list(experiments.identifiers())}")

    experiments.as_file(params.output.experiments)
    joint_table = flex.reflection_table()
    for reflection_table in reflections:
        joint_table.extend(reflection_table)
    joint_table.as_file(params.output.reflections)


if __name__ == "__main__":
    run()



"""
Print a table of the orientation for every image of a dataset. The
orientation is expressed as a zone axis (a direction referenced to the direct
lattice) [uvw] giving the beam direction with respect to the crystal lattice.
Take into account any scan-varying models.

Usage: dials.frame_orientations refined.expt
"""


from __future__ import annotations

import sys
from itertools import tee

import matplotlib

from scitbx import matrix

import dials.util
from dials.util import tabulate
from dials.util.options import ArgumentParser, flatten_experiments

matplotlib.use("Agg")
import matplotlib.pyplot as plt


# An itertools recipe in Python 3.7, but a module function in 3.10
def pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


class Script:
    """A class for running the script."""

    def __init__(self):
        """Initialise the script."""
        from libtbx.phil import parse

        # The phil scope
        phil_scope = parse(
            """
scale = unit *max_cell ewald_sphere_radius
    .type = choice
    .help = "Choose the scale for the direction vector in orthogonal"
            "coordinates prior to transformation into fractional"
            "coordinates [uvw]"

plot_filename = None
    .type = str
    .help = "Filename for a plot of angle between neighbouring frames"
            "(set to None for no plot)"
""",
            process_includes=True,
        )

        usage = "dials.frame_orientations refined.expt refined.refl"

        # Create the parser
        self.parser = ArgumentParser(
            usage=usage,
            phil=phil_scope,
            read_experiments=True,
            check_format=False,
            epilog=__doc__,
        )

    def run(self, args=None):
        """Execute the script."""

        # Parse the command line
        self.params, _ = self.parser.parse_args(args, show_diff_phil=True)

        if not self.params.input.experiments:
            self.parser.print_help()
            sys.exit()

        # Try to load the models
        experiments = flatten_experiments(self.params.input.experiments)
        nexp = len(experiments)
        if nexp == 0:
            self.parser.print_help()
            sys.exit("No Experiments found in the input")

        # Set up a plot if requested
        if self.params.plot_filename:
            plt.figure()

        header = [
            "Image",
            "Beam direction (xyz)",
            "Zone axis [uvw]",
            "Angles between beam\nand axes a, b, c (deg)",
            "Angle from\nprevious image (deg)",
        ]
        for iexp, exp in enumerate(experiments):
            print(f"For Experiment id = {iexp}")
            print(exp.beam)
            print(exp.crystal)
            print(exp.scan)

            if self.params.scale == "ewald_sphere_radius":
                scale = 1.0 / exp.beam.get_wavelength()
            elif self.params.scale == "max_cell":
                uc = exp.crystal.get_unit_cell()
                scale = max(uc.parameters()[0:3])
            else:
                scale = 1.0
            print(
                "Beam direction scaled by {} = {:.3f} to "
                "calculate zone axis\n".format(self.params.scale, scale)
            )

            dat = extract_experiment_data(exp, scale)
            images = dat["images"]
            directions = dat["directions"]
            zone_axes = dat["zone_axes"]
            real_space_axes = dat["real_space_axes"]

            # calculate the angle between the beam and each crystal axis
            axis_angles = []
            for d, rsa in zip(directions, real_space_axes):
                angles = [d.angle(a, deg=True) for a in rsa]
                axis_angles.append("{:.2f} {:.2f} {:.2f}".format(*angles))

            # calculate the orientation offset between each image
            offset = [
                e1.angle(e2, deg=True) for e1, e2 in zip(zone_axes[:-1], zone_axes[1:])
            ]
            str_off = ["---"] + [f"{e:.8f}" for e in offset]

            rows = []
            for i, d, z, a, o in zip(
                images,
                directions,
                zone_axes,
                axis_angles,
                str_off,
            ):
                row = [
                    str(i),
                    "{:.8f} {:.8f} {:.8f}".format(*d.elems),
                    "{:.8f} {:.8f} {:.8f}".format(*z.elems),
                    a,
                    o,
                ]
                rows.append(row)

            # Print the table
            print(tabulate(rows, header))

            # Add to the plot, if requested
            if self.params.plot_filename:
                plt.scatter(images[1:], offset, s=1)

        # Finish and save plot, if requested
        if self.params.plot_filename:
            plt.xlabel("Image number")
            plt.ylabel(r"Angle from previous image $\left(^\circ\right)$")
            plt.title(r"Angle between neighbouring images")
            print(f"Saving plot to {self.params.plot_filename}")
            plt.savefig(self.params.plot_filename)

        print()


def extract_experiment_data(exp, scale=1):
    """Extract lists of the image number, beam direction and zone axis from an
    experiment"""
    crystal = exp.crystal
    beam = exp.beam
    scan = exp.scan
    gonio = exp.goniometer

    image_range = scan.get_image_range()
    images = list(range(image_range[0], image_range[1] + 1))
    num_scan_points = scan.get_num_images() + 1

    if beam.num_scan_points > 0:
        us0 = []
        for i in range(beam.num_scan_points):
            s0 = matrix.col(beam.get_s0_at_scan_point(i))
            us0.append(s0.normalize())
    else:
        us0 = [matrix.col(beam.get_unit_s0()) for _ in range(num_scan_points)]

    if gonio.num_scan_points > 0:
        S_mats = [
            matrix.sqr(gonio.get_setting_rotation_at_scan_point(i))
            for i in range(gonio.num_scan_points)
        ]
    else:
        S_mats = [
            matrix.sqr(gonio.get_setting_rotation()) for _ in range(num_scan_points)
        ]

    F_mats = [matrix.sqr(gonio.get_fixed_rotation()) for _ in range(num_scan_points)]
    start, stop = scan.get_array_range()
    R_mats = []
    axis = matrix.col(gonio.get_rotation_axis_datum())
    for i in range(start, stop + 1):
        phi = scan.get_angle_from_array_index(i, deg=False)
        R = matrix.sqr(axis.axis_and_angle_as_r3_rotation_matrix(phi, deg=False))
        R_mats.append(R)

    if crystal.num_scan_points > 0:
        U_mats = [
            matrix.sqr(crystal.get_U_at_scan_point(i))
            for i in range(crystal.num_scan_points)
        ]
        B_mats = [
            matrix.sqr(crystal.get_B_at_scan_point(i))
            for i in range(crystal.num_scan_points)
        ]
    else:
        U_mats = [matrix.sqr(crystal.get_U()) for _ in range(num_scan_points)]
        B_mats = [matrix.sqr(crystal.get_B()) for _ in range(num_scan_points)]

    check = {len(x) for x in (us0, S_mats, F_mats, R_mats, U_mats)}
    assert len(check) == 1
    assert check.pop() == len(images) + 1

    # Construct full orientation matrix in the lab frame for each scan-point
    SRFU = (S * R * F * U for S, R, F, U in zip(S_mats, R_mats, F_mats, U_mats))

    # Now convert this to the orientation matrix at the centre of each frame by
    # calculating the linear transform that goes from the start of the frame
    # to the end, and then applying half of that to the start orientation
    U_frames = []
    for U1, U2 in pairwise(SRFU):
        M = U2 * U1.transpose()
        (
            angle,
            axis,
        ) = M.r3_rotation_matrix_as_unit_quaternion().unit_quaternion_as_axis_and_angle(
            deg=False
        )
        M_half = axis.axis_and_angle_as_r3_rotation_matrix(angle / 2, deg=False)
        U_frames.append(M_half * U1)

    # Convert the crystal B matrix at scan-points into a B matrix at the frame
    # centres. In this case, the transformation is not a rotation. Approximate
    # the answer by taking the average of the two B matrices.
    # FIXME is this kosher?
    B_frames = []
    for B1, B2 in pairwise(B_mats):
        B_frames.append((B1 + B2) / 2)

    UB_frames = [U * B for U, B in zip(U_frames, B_frames)]
    # This is the orthogonalisation matrix for the reciprocal space laboratory
    # frame. We want the real space fractionalisation matrix, which is its
    # transpose (https://dials.github.io/documentation/conventions.html)
    frac_mats = [m.transpose() for m in UB_frames]

    # Calculate zone axes, which also requires the beam directions at the frame
    # centres
    us0_frames = []
    for d1, d2 in pairwise(us0):
        us0_frames.append(((d1 + d2) / 2).normalize())
    zone_axes = [frac * (d * scale) for frac, d in zip(frac_mats, us0_frames)]

    # Now get the real space orthogonalisation matrix to calculate the real
    # space cell vectors at each image
    orthog_mats = (frac.inverse() for frac in frac_mats)
    h = matrix.col((1, 0, 0))
    k = matrix.col((0, 1, 0))
    l = matrix.col((0, 0, 1))
    real_space_axes = [(o * h, o * k, o * l) for o in orthog_mats]
    return {
        "images": images,
        "directions": us0_frames,
        "zone_axes": zone_axes,
        "real_space_axes": real_space_axes,
        "orientations": U_frames,
    }


@dials.util.show_mail_handle_errors()
def run(args=None):
    script = Script()
    script.run(args)


if __name__ == "__main__":
    run()



"""
Optimise the rotation axis orientation using the method from Gorelik et al.
(https://doi.org/10.1007/978-94-007-5580-2_31) with code adapted from
Stef Smeets's edtools (https://github.com/instamatic-dev/edtools).

Three levels of search can be performed: a global search, a local search and a
fine search. At each trial position the sharpness of a cylindrical projection of
reciprocal lattice points is assessed via its variance. The correct orientation
maximises this variance.

Examples::

  dials.find_rotation_axis imported.expt strong.refl
"""

from __future__ import annotations

import concurrent.futures
import logging
import sys

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

import dxtbx.flumpy as flumpy
import libtbx
import libtbx.phil

import dials.util
import dials.util.log
from dials.array_family import flex
from dials.util.mp import available_cores

# from dials.array_family import flex
from dials.util.options import ArgumentParser, reflections_and_experiments_from_files
from dials.util.version import dials_version

# Define a logger
logger = logging.getLogger("dials.find_rotation_axis")

# Define the master PHIL scope for this program.
phil_scope = libtbx.phil.parse(
    """
max_two_theta = 10.0
    .type = float
    .help = "Scattering angle limit to select reflections only in the central"
            "mostly flat region of the Ewald sphere surface"

view = False
    .type = bool
    .help = "View phi/theta histogram with current rotation axis (azimuth)"
            "instead of performing an optimisation"

global_search = True
    .type = bool
    .help = "Perform global search of the azimuthal angle. If False, only a"
            "local search will be performed around the start value or the value"
            "given by azimuth=VAL"

azimuth = None
    .type = float
    .help = "Use the given value of azimuth to plot the histogram or as the"
            "starting point for the optimization"

opposite = False
    .type = bool
    .help = "Try the opposite from the initial value (or that given by"
            "azimuth=VAL)"

nproc = Auto
    .help = "Number of processes over which to split the search. If set to"
            "Auto, DIALS will choose automatically."
    .type = int(value_min=1)
    .expert_level = 1

output {
    experiments = optimised.expt
        .type = path

    log = "dials.find_rotation_axis.log"
        .type = str

    plot = find_rotation_axis
        .type = str
        .help = "Filename prefix for plots. Set to None for no plots"
}
"""
)


def rotation_axis_to_xyz(rotation_axis, invert=False):
    """
    Convert rotation axis angle to XYZ vector compatible with DIALS. Set invert
    to 'True' for anti-clockwise rotation
    """
    if invert:
        rotation_axis += np.pi

    rot_x = np.cos(rotation_axis)
    rot_y = np.sin(rotation_axis)
    rot_z = 0

    return rot_x, -rot_y, rot_z


def rotation_matrix(axis, theta):
    """Calculates the rotation matrix around axis of angle theta (radians)"""

    l = np.sqrt(np.dot(axis, axis))
    axis = axis / l

    a = np.cos(theta / 2)
    b, c, d = -1 * axis * np.sin(theta / 2)

    return np.array(
        [
            [a * a + b * b - c * c - d * d, 2 * (b * c - a * d), 2 * (b * d + a * c)],
            [2 * (b * c + a * d), a * a + c * c - b * b - d * d, 2 * (c * d - a * b)],
            [2 * (b * d - a * c), 2 * (c * d + a * b), a * a + d * d - b * b - c * c],
        ]
    )


def make_2d_rotmat(theta):
    """Take angle in radians, and return 2D rotation matrix"""
    R = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])
    return R


def random_sample(arr, n):
    """Select random sample of `n` rows from array"""
    indices = np.random.choice(arr.shape[0], n, replace=False)
    return arr[indices]


def xyz2cyl(arr):
    """
    Take a set of reflections in XYZ and convert to polar (cylindrical)
    coordinates
    """
    sx, sy, sz = arr.T
    out = np.empty((len(arr), 2))
    np.hypot(sx, sy, out=out[:, 0])
    np.arctan2(sz, out[:, 0], out=out[:, 1])
    np.arctan2(sy, sx, out=out[:, 0])
    return out


def cylinder_histo(xyz, bins=(1000, 500)):
    """
    Take reciprocal lattice vectors in XYZ format and output cylindrical
    projection. `bins` gives the resolution of the 2D histogram.
    """
    i, j = np.triu_indices(len(xyz), k=1)
    diffs = xyz[i] - xyz[j]
    polar = xyz2cyl(diffs)

    px, py = polar.T
    H, xedges, yedges = np.histogram2d(
        px,
        py,
        bins=bins,
        range=[
            [
                -np.pi,
                np.pi,
            ],
            [-np.pi / 2, np.pi / 2],
        ],
    )

    return H, xedges, yedges


def plot_histo(H, xedges, yedges, title="Histogram", filename=None):
    """Plot the histogram of the cylindrical projection."""
    plt.figure(figsize=(10, 7))
    plt.imshow(
        H.T,
        interpolation="nearest",
        origin="lower",
        extent=[xedges[0], xedges[-1], yedges[0], yedges[-1]],
        vmax=np.percentile(H, 99),
    )
    plt.title(title)
    plt.xlim(-np.pi, np.pi)
    plt.ylim(-np.pi / 2, np.pi / 2)
    plt.xlabel("phi ($\\pi$)")
    plt.ylabel("theta ($\\pi$)")
    if filename:
        plt.savefig(filename)
    else:
        plt.show()


def make(arr, azimuth: float, wavelength: float):
    """
    Prepare xyz (reciprocal space coordinates) from reflection positions/angle
    (`arr`). The azimuth angle (in degrees) gives the orientation of the
    rotation axis in the plane of the image, defined as the angle between x
    (horizontal axis pointing right) and the rotation axis going in a clockwise
    direction
    """

    reflections = arr[:, 0:2]
    angle = arr[:, 2]

    azimuth_rad = np.radians(azimuth)
    r = make_2d_rotmat(azimuth_rad)

    refs_ = np.dot(reflections, r)

    y, x = refs_.T

    R = 1 / wavelength
    C = R - np.sqrt(R**2 - x**2 - y**2).reshape(-1, 1)
    xyz = (
        np.c_[x * np.cos(angle), y, -x * np.sin(angle)]
        + C * np.c_[-np.sin(angle), np.zeros_like(angle), -np.cos(angle)]
    )

    return xyz


def calc_var(arr, azimuth, wavelength, hist_bins):
    xyz = make(arr, azimuth, wavelength)
    H, _, _ = cylinder_histo(xyz, bins=hist_bins)
    var = np.var(H)
    return var


def optimise(
    arr,
    azimuth_start: float,
    wavelength=float,
    plusminus: int = 180,
    step: int = 10,
    hist_bins: (int, int) = (1000, 500),
    nproc: int = 1,
) -> float:
    """
    Optimise the value of azimuth around the given point.

    Args:
        azimuth_start: defines the starting angle
        step, plusminus: together with azimuth_start define the range of values
            to loop over
        hist_bins: size of the 2d histogram to produce the final phi/theta plot
        nproc: number of processes over which to split the search

    Returns:
        The best value for the azimuth angle
    """

    r = np.arange(azimuth_start - plusminus, azimuth_start + plusminus, step)

    best_score = 0
    best_azimuth = 0

    if nproc > 1:
        with concurrent.futures.ProcessPoolExecutor(max_workers=nproc) as pool:
            variances = [
                pool.submit(calc_var, arr, azimuth, wavelength, hist_bins)
                for azimuth in r
            ]
        variances = [e.result() for e in variances]
    else:
        # For nproc=1 keep the jobs in the main process
        variances = [calc_var(arr, azimuth, wavelength, hist_bins) for azimuth in r]

    for azimuth, var in zip(r, variances):
        logger.info(f"azimuth: {azimuth:8.2f}, variance: {var:5.2f}")

        # Collate data for the rotation axis distribution curve
        xvals.append(azimuth)
        vvals.append(var)

        if var > best_score:
            best_azimuth = azimuth
            best_score = var

    logger.info(f"Best azimuth: {best_azimuth:.2f}; score: {best_score:.2f}")

    return best_azimuth


def extract_spot_data(reflections, experiments, max_two_theta):
    """
    From the spot positions, extract reciprocal space X, Y and angle positions
    for each reflection up to the scattering angle max_two_theta
    """
    # Map reflections to reciprocal space
    reflections.centroid_px_to_mm(experiments)

    # Calculate scattering vectors
    reflections["s1"] = flex.vec3_double(len(reflections))
    reflections["2theta"] = flex.double(len(reflections))
    panel_numbers = flex.size_t(reflections["panel"])
    for i, expt in enumerate(experiments):
        if "imageset_id" in reflections:
            sel_expt = reflections["imageset_id"] == i
        else:
            sel_expt = reflections["id"] == i

        for i_panel in range(len(expt.detector)):
            sel = sel_expt & (panel_numbers == i_panel)
            x, y, _ = reflections["xyzobs.mm.value"].select(sel).parts()
            s1 = expt.detector[i_panel].get_lab_coord(flex.vec2_double(x, y))
            s1 = s1 / s1.norms() * (1 / expt.beam.get_wavelength())
            tt = s1.angle(expt.beam.get_s0(), deg=True)
            reflections["s1"].set_selected(sel, s1)
            reflections["2theta"].set_selected(sel, tt)

    # Filter reflections
    full_len = len(reflections)
    reflections = reflections.select(reflections["2theta"] <= max_two_theta)
    if len(reflections) < full_len:
        logger.info(
            f"{len(reflections)} reflections with 2θ ≤ {max_two_theta}° selected from {full_len} total"
        )

    x, y, _ = reflections["s1"].parts()
    _, _, angle = reflections["xyzobs.mm.value"].parts()
    arr = flumpy.to_numpy(x)
    arr = np.c_[
        x, flumpy.to_numpy(-y)
    ]  # Y is inverted to match calculation in edtools.find_rotation_axis
    arr = np.c_[arr, flumpy.to_numpy(angle)]
    return arr


@dials.util.show_mail_on_error()
def run(args=None, phil=phil_scope):
    """
    Load diffraction geometry and spot positions, optimise rotation axis
    orientation and save output files as specified.

    Args:
        args (list): Additional command-line arguments
        phil: The working PHIL parameters

    Returns:
        None
    """

    usage = "dials.find_rotation_axis [options] imported.expt strong.refl"

    parser = ArgumentParser(
        usage=usage,
        phil=phil,
        read_reflections=True,
        read_experiments=True,
        check_format=False,
        epilog=__doc__,
    )

    params, options = parser.parse_args(args=args, show_diff_phil=False)

    # Configure the logging.
    dials.util.log.config(options.verbose, logfile=params.output.log)

    # Log the dials version
    logger.info(dials_version())

    # Log the difference between the PHIL scope definition and the active PHIL
    # scope, which will include the parsed user inputs.
    diff_phil = parser.diff_phil.as_str()
    if diff_phil:
        logger.info("The following parameters have been modified:\n%s", diff_phil)

    reflections, experiments = reflections_and_experiments_from_files(
        params.input.reflections, params.input.experiments
    )

    # Check the models and data
    nexp = len(experiments)
    if nexp > 1:
        logger.info(
            "Only the first experiment will be used to determine oscillation and current rotation axis"
        )
        if len(experiments.goniometers()) > 1:
            sys.exit("Multiple experiments must share a goniometer model")
    if nexp == 0 or len(reflections) == 0:
        parser.print_help()
        return
    if len(reflections) > 1:
        sys.exit("Only one reflections list can be imported at present")
    reflections = reflections[0]
    expt = experiments[0]
    wavelength = expt.beam.get_wavelength()
    rotx, roty, _ = expt.goniometer.get_rotation_axis()
    azimuth_current = np.degrees(np.arctan2(-roty, rotx))
    arr = extract_spot_data(reflections, experiments, params.max_two_theta)

    if params.azimuth is not None:
        azimuth_current = params.azimuth

    azimuth_opposite = azimuth_current + 180

    if params.opposite:
        azimuth_current = azimuth_opposite

    if azimuth_current > 180:
        azimuth_current -= 360

    if azimuth_opposite > 180:
        azimuth_opposite -= 360

    logger.info(
        f"Wavelength: {wavelength:.5f} Ångström\n"
        f"Azimuth (current): {azimuth_current:.5f} degrees\n"
        f"                 {np.radians(azimuth_current):.5f} radians"
    )

    if params.nproc is libtbx.Auto:
        params.nproc = available_cores()
        logger.info("Setting nproc={}".format(params.nproc))

    hist_bins = 1000, 500

    if params.view:
        azimuth_final = azimuth_current
    else:
        global xvals
        global vvals
        xvals = []
        vvals = []

        azimuth_global = azimuth_local = azimuth_fine = 0

        if not params.global_search:
            azimuth_tmp = azimuth_global = azimuth_current
        else:
            logger.info("Performing global search")
            azimuth_tmp = 0
            azimuth_global = azimuth_tmp = optimise(
                arr,
                azimuth_tmp,
                wavelength,
                plusminus=180,
                step=5,
                hist_bins=hist_bins,
                nproc=params.nproc,
            )

        logger.info("Performing local search")
        azimuth_local = azimuth_tmp = optimise(
            arr,
            azimuth_tmp,
            wavelength,
            plusminus=5,
            step=1,
            hist_bins=hist_bins,
            nproc=params.nproc,
        )

        logger.info("Performing fine search")
        azimuth_fine = azimuth_tmp = optimise(
            arr,
            azimuth_tmp,
            wavelength,
            plusminus=1,
            step=0.1,
            hist_bins=hist_bins,
            nproc=params.nproc,
        )

        azimuth_final = azimuth_tmp

        logger.info(
            "---\n"
            f"Best azimuth (global search): {azimuth_global:.3f}\n"
            f"Best azimuth (local search): {azimuth_local:.3f}\n"
            f"Best azimuth (fine search): {azimuth_fine:.3f}"
        )

    xyz = make(arr, azimuth_final, wavelength)
    H, xedges, yedges = cylinder_histo(xyz)

    var = np.var(H)
    logger.info(f"Variance: {var:.2f}")

    # check opposite
    xyz_opp = make(arr, azimuth_final + 180, wavelength)
    H_opp, xedges_opp, yedges_opp = cylinder_histo(xyz_opp)

    var_opp = np.var(H_opp)
    logger.info(f"Variance (opposite): {var_opp:.2f}")

    if var < var_opp:
        logger.info(
            f"\nOpposite angle ({azimuth_opposite:.2f}°) has higher variance!\n"
        )

    if params.output.plot:
        matplotlib.use("Agg")
        hist_filename = params.output.plot + "-histogram.png"
        plot_histo(
            H,
            xedges,
            yedges,
            title=f"azimuth={azimuth_final:.2f}$^\\circ$ | var={var:.2f}",
            filename=hist_filename,
        )

        if not params.view:
            # Plot rotation axis distribution curve
            plt.figure(figsize=(10, 7))
            plt.scatter(xvals, vvals, marker="+", lw=1.0, color="red")
            plt.xlabel("Rotation axis position ($^\\circ$)")
            plt.ylabel("Variance of the polar coordinate histogram")
            plt.title(
                f"Rotation axis determination | Maximum @ {azimuth_final:.2f}$^\\circ$"
            )
            projection_filename = params.output.plot + "-projection.png"
            plt.savefig(projection_filename)

    azimuth_deg = azimuth_final
    azimuth_rad = np.radians(azimuth_final)

    logger.info(
        f"\nRotation axis found: {azimuth_deg:.2f} deg. / {azimuth_rad:.3f} rad."
    )
    expt.goniometer.set_rotation_axis(rotation_axis_to_xyz(azimuth_rad))
    logger.info(str(expt.goniometer))

    logger.info(f"Saving optimised experiments to {params.output.experiments}")
    experiments.as_file(params.output.experiments)


if __name__ == "__main__":
    run()



"""Tests for dials.compute_delta_cchalf."""

from __future__ import annotations

import os

import procrunner
import pytest

from dials.command_line.compute_delta_cchalf import CCHalfFromMTZ, phil_scope, run


def check_cchalf_result(fileobj):
    """Inspect the result"""
    lines = fileobj.readlines()
    assert lines[0] == "1 -0.004673\n"
    assert lines[1] == "0 0.001234\n"


def test_suitable_exit_on_bad_input(dials_data, run_in_tmp_path):
    location = dials_data("l_cysteine_4_sweeps_scaled", pathlib=True)
    refl = location / "scaled_35.refl"
    expt = location / "scaled_35.expt"

    args = [str(refl), str(expt)]
    with pytest.raises(SystemExit):
        run(args)

    args = [str(refl), str(expt), "mode=image_group", "group_size=10000"]
    with pytest.raises(SystemExit):
        run(args)


def test_compute_delta_cchalf_scaled_data(dials_data, tmp_path):
    """Test dials.compute_delta_cchalf on scaled data."""
    location = dials_data("l_cysteine_4_sweeps_scaled", pathlib=True)
    refls = location / "scaled_20_25.refl"
    expts = location / "scaled_20_25.expt"

    # set cutoff to 0.0 to force one to be 'rejected'
    command = [
        "dials.compute_delta_cchalf",
        refls,
        expts,
        "stdcutoff=0.0",
        "output.reflections=filtered.refl",
        "output.experiments=filtered.expt",
    ]
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / "filtered.expt").is_file()
    assert (tmp_path / "filtered.refl").is_file()
    assert (tmp_path / "delta_cchalf.dat").is_file()
    assert (tmp_path / "compute_delta_cchalf.html").is_file()
    with open(tmp_path / "delta_cchalf.dat") as f:
        check_cchalf_result(f)


def test_compute_delta_cchalf_scaled_data_mtz(dials_data, tmp_path):
    """Test dials.compute_delta_cchalf on scaled data."""
    location = dials_data("l_cysteine_4_sweeps_scaled", pathlib=True)
    refls = location / "scaled_20_25.refl"
    expts = location / "scaled_20_25.expt"

    # First export the data
    command = ["dials.export", refls, expts, "partiality_threshold=0.99"]
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / "scaled.mtz").is_file()

    # set cutoff to 0.0 to force one to be 'rejected'
    command = [
        "dials.compute_delta_cchalf",
        f"mtzfile={tmp_path / 'scaled.mtz'}",
        "stdcutoff=0.0",
    ]
    result = procrunner.run(command, working_directory=tmp_path)
    assert not result.returncode and not result.stderr
    assert (tmp_path / "delta_cchalf.dat").is_file()
    assert (tmp_path / "compute_delta_cchalf.html").is_file()
    with open(tmp_path / "delta_cchalf.dat") as f:
        check_cchalf_result(f)


def test_compute_delta_cchalf(dials_regression):
    """Test compute delta cchalf on an integrated mtz."""

    filename = os.path.join(
        dials_regression, "delta_cchalf_test_data", "test.XDS_ASCII.mtz"
    )
    params = phil_scope.extract()
    params.nbins = 1

    script = CCHalfFromMTZ(params, filename)
    script.run()

    cchalf_i = script.results_summary["per_dataset_delta_cc_half_values"][
        "delta_cc_half_values"
    ]

    assert abs(100 * script.results_summary["mean_cc_half"] - 94.582) < 1e-3
    assert (
        abs(100 * script.results_summary["mean_cc_half"] - 100 * cchalf_i[1] - 79.587)
        < 1e-3
    )
    assert (
        abs(100 * script.results_summary["mean_cc_half"] - 100 * cchalf_i[0] - 94.238)
        < 1e-3
    )



from __future__ import annotations

import os

import procrunner
import pytest

cpp_tests = [
    # Paths are under /build/
    "tests/algorithms/spatial_indexing/tst_collision_detection",
    "tests/algorithms/spot_prediction/tst_reeke_model",
]


@pytest.mark.parametrize(
    "executable", cpp_tests, ids=[p.replace("/", "-") for p in cpp_tests]
)
def test_cpp_program(executable):
    if "LIBTBX_BUILD" not in os.environ:
        pytest.skip("LIBTBX_ENV is unset; don't know how to find test executable")
    full_path = os.path.join(
        os.environ["LIBTBX_BUILD"], "dials", *(executable.split("/"))
    )
    print(full_path)

    result = procrunner.run([full_path])
    assert not result.returncode and not result.stderr



from __future__ import annotations

import logging

logger = logging.getLogger(__name__)


class MaskerBase:
    """A root class to that does overlap masking"""

    def __init__(self, experiment):
        """Initialise the overlap masking algorithm

        Params:
            experiment The experiment data
        """
        from dials.algorithms.shoebox import MaskOverlapping

        # Construct the overlapping reflection mask
        self.mask_overlapping = MaskOverlapping()

    def __call__(self, reflections, adjacency_list=None):
        """Mask the given reflections.

        Params:
            reflections The reflection list
            adjacency_list The adjacency_list (optional)

        Returns:
            The masked reflection list
        """

        # Mask the overlaps if an adjacency list is given
        if adjacency_list:
            logger.info("Masking overlapping reflections")
            self.mask_overlapping(
                reflections["shoebox"], reflections["xyzcal.px"], adjacency_list
            )
            logger.info("Masked %s overlapping reflections", len(adjacency_list))

        # Return the reflections
        return reflections


class Masker3DProfile(MaskerBase):
    """A class to perform 3D profile masking"""

    def __init__(self, experiments):
        """Initialise the masking algorithms

        Params:
            experiment The experiment data
            delta_d The extent of the reflection in reciprocal space
            delta_m The extent of the reflection in reciprocal space
        """
        super().__init__(experiments[0])
        self._experiments = experiments

    def __call__(self, reflections, adjacency_list=None):
        """Mask the given reflections.

        Params:
            reflections The reflection list
            adjacency_list The adjacency_list (optional)

        Returns:
            The masked reflection list
        """
        reflections = super().__call__(reflections, adjacency_list)

        # Mask the foreground region
        self._experiment.profile.compute_mask(
            reflections,
            self._experiment.crystal,
            self._experiment.beam,
            self._experiment.detector,
            self._experiment.goniometer,
            self._experiment.scan,
        )

        # Return the reflections
        return reflections


class MaskerEmpirical(MaskerBase):
    """A class to perform empirical masking"""

    def __init__(self, experiment, reference):
        """Initialise the masking algorithms

        Params:
            experiment The experiment data
        """
        super().__init__(experiment)

        from dials.algorithms.shoebox import MaskEmpirical

        # Construct the foreground pixel mask
        self.mask_empirical = MaskEmpirical(reference)
        self._reference = reference

    def __call__(self, reflections, adjacency_list=None):
        """Mask the given reflections.

        Params:
            reflections The reflection list
            adjacency_list The adjacency_list (optional)

        Returns:
            The masked reflection list
        """
        reflections = super().__call__(reflections, adjacency_list)

        if self.mask_empirical:
            # Mask the foreground region
            self.mask_empirical(reflections)

        # Return the reflections
        return reflections
